from typing import Dict

from verbecc.src.defs.types.mood import Mood
from verbecc.src.defs.types.tense import Tense
from verbecc.src.defs.types.lang_code import LangCodeISO639_1

MOOD_MAP: Dict[Mood, Dict[LangCodeISO639_1, Mood]] = {
    Mood.en.Indicative: {
        LangCodeISO639_1.ca: Mood.ca.Indicatiu,
        LangCodeISO639_1.es: Mood.es.Indicativo,
        LangCodeISO639_1.fr: Mood.fr.Indicatif,
        LangCodeISO639_1.it: Mood.it.Indicativo,
        LangCodeISO639_1.pt: Mood.pt.Indicativo,
        LangCodeISO639_1.ro: Mood.ro.Indicativ,
    },
    Mood.en.Subjunctive: {
        LangCodeISO639_1.ca: Mood.ca.Subjuntiu,
        LangCodeISO639_1.es: Mood.es.Subjuntivo,
        LangCodeISO639_1.fr: Mood.fr.Subjonctif,
        LangCodeISO639_1.it: Mood.it.Congiuntivo,
        LangCodeISO639_1.pt: Mood.pt.Subjuntivo,
        LangCodeISO639_1.ro: Mood.ro.Subjunctiv,
    },
    Mood.en.Imperative: {
        LangCodeISO639_1.ca: Mood.ca.Imperatiu,
        LangCodeISO639_1.es: Mood.es.Imperativo,
        LangCodeISO639_1.fr: Mood.fr.Imperatif,
        LangCodeISO639_1.it: Mood.it.Imperativo,
        LangCodeISO639_1.pt: Mood.pt.Imperativo,
        LangCodeISO639_1.ro: Mood.ro.Imperativ,
    },
    Mood.en.Conditional: {
        LangCodeISO639_1.ca: Mood.ca.Condicional,
        LangCodeISO639_1.es: Mood.es.Condicional,
        LangCodeISO639_1.fr: Mood.fr.Conditionnel,
        LangCodeISO639_1.it: Mood.it.Condizionale,
        LangCodeISO639_1.pt: Mood.pt.Condicional,
        LangCodeISO639_1.ro: "<not implemented>",  # The conditional mood is formed by combining the conditional particle "ar" with the infinitive form of the verb
    },
    Mood.en.Infinitive: {
        LangCodeISO639_1.ca: Mood.ca.Infinitiu,
        LangCodeISO639_1.es: Mood.es.Infinitivo,
        LangCodeISO639_1.fr: Mood.fr.Infinitif,
        LangCodeISO639_1.it: Mood.it.Infinito,
        LangCodeISO639_1.pt: Mood.pt.Infinitivo,
        LangCodeISO639_1.ro: Mood.ro.Infinitiv,
    },
    # TODO: Make French consistent with other languages i.e.
    # make gerund and past-participle different moods i.e.
    # participe.participe-présent -> participe-présent.participe
    # participe.participe-passé -> participe-passé.participe
    Mood.en.Gerund: {
        LangCodeISO639_1.ca: Mood.ca.Gerundi,  # gerundi.gerundi
        LangCodeISO639_1.es: Mood.es.Gerundio,  # gerundio.gerundio
        LangCodeISO639_1.fr: Mood.fr.Participe,  # participe.participe-présent is gerund in French
        LangCodeISO639_1.it: Mood.it.Infinito,  # infinito.gerundio is gerund in Italian
        LangCodeISO639_1.pt: Mood.pt.Gerúndio,  # gerúndio.gerúndio
        LangCodeISO639_1.ro: Mood.ro.Gerunziu,  # gerunziu.gerunziu
    },
    Mood.en.Participle: {
        LangCodeISO639_1.ca: Mood.ca.Participi,  # particip.particip
        LangCodeISO639_1.es: Mood.es.Participo,  # participo.participo
        LangCodeISO639_1.fr: Mood.fr.Participe,  # participe.participe-présent, participe.participe-passé
        LangCodeISO639_1.it: Mood.it.Participio,  # participio.participio
        LangCodeISO639_1.pt: Mood.pt.Particípio,  # particípio.particípio
        LangCodeISO639_1.ro: Mood.ro.Participiu,  # participiu.participiu
    },
}

TENSE_MAP = {
    Tense.en.Present: {
        LangCodeISO639_1.ca: Tense.ca.Present,
        LangCodeISO639_1.es: Tense.es.Presente,
        LangCodeISO639_1.fr: Tense.fr.Présent,
        LangCodeISO639_1.it: Tense.it.Presente,
        LangCodeISO639_1.pt: Tense.pt.Presente,
        LangCodeISO639_1.ro: Tense.ro.Prezent,
    },
    Tense.en.Imperfect: {
        LangCodeISO639_1.ca: Tense.ca.Imperfet,
        LangCodeISO639_1.es: Tense.es.PretéritoImperfecto,
        LangCodeISO639_1.fr: Tense.fr.Imparfait,
        LangCodeISO639_1.it: Tense.it.Imperfetto,
        LangCodeISO639_1.pt: Tense.pt.Imperfeito,
    },
    Tense.en.SimplePast: {
        LangCodeISO639_1.ca: Tense.ca.PassatSimple,
        LangCodeISO639_1.es: Tense.es.PretéritoPerfectoSimple,
        LangCodeISO639_1.fr: Tense.fr.PasséSimple,
        LangCodeISO639_1.it: Tense.it.PassatoRemoto,  # The passato prossimo (compound tense) is roughly equivalent to the present perfect and past simple tenses
        # however passato-remoto is the literal simple past tense e.g. 'io fui, tu fosti, lui fu, noi fummo',
        LangCodeISO639_1.pt: Tense.pt.PretéritoPerfeito,
        LangCodeISO639_1.ro: Tense.ro.PerfectSimplu,
    },
    Tense.en.Future: {
        LangCodeISO639_1.ca: Tense.ca.Futur,
        LangCodeISO639_1.es: Tense.es.Futuro,
        LangCodeISO639_1.fr: Tense.fr.FuturSimple,
        LangCodeISO639_1.it: Tense.it.Futuro,
        LangCodeISO639_1.pt: Tense.pt.FuturoDoPresente,
        LangCodeISO639_1.ro: Tense.ro.Viitor1,  # Romanian has viitor-1, viitor-2, viitor-1-popular, viitor-2-popular
    },
    Tense.en.Gerund: {
        LangCodeISO639_1.ca: Tense.ca.Gerundi,
        LangCodeISO639_1.es: Tense.es.Gerundio,
        LangCodeISO639_1.fr: Tense.fr.ParticipePresent,  # TODO: Make French consistent
        LangCodeISO639_1.it: Tense.it.Gerundio,
        LangCodeISO639_1.pt: Tense.pt.Gerúndio,
        LangCodeISO639_1.ro: Tense.ro.Gerunziu,
    },
    Tense.en.PastParticiple: {
        LangCodeISO639_1.ca: Tense.ca.Particip,
        LangCodeISO639_1.es: Tense.es.Participo,
        LangCodeISO639_1.fr: Tense.fr.ParticipePassé,
        LangCodeISO639_1.it: Tense.it.ParticipioPassato,
        LangCodeISO639_1.pt: Tense.pt.Particípio,
        LangCodeISO639_1.ro: Tense.ro.Participiu,
    },
}


def xmood(lang: LangCodeISO639_1, m: Mood) -> Mood:
    """Takes a mood name in EN and translates it to the specified language

    :raises: Exception if mood or lang doesn't exist
    """
    return MOOD_MAP[m][lang]


def xtense(lang: LangCodeISO639_1, t: Tense) -> Tense:
    """Takes a tense name in EN and translates it to the specified language

    :raises: Exception if tense or lang doesn't exist
    """
    return TENSE_MAP[t][lang]



from lxml import etree

from verbecc.src.parsers.tense_template import TenseTemplate
from verbecc.src.defs.types.lang_code import LangCodeISO639_1 as Lang
from verbecc.src.defs.types.tense import Tense


def test_tense_and_person():
    mood = "indicatif"
    tense_elem_str: str = """<présent>
        <p><i>ie</i><i>ye</i></p>
        <p><i>ies</i><i>yes</i></p>
        <p><i>ie</i><i>ye</i></p>
        <p><i>yons</i></p>
        <p><i>yez</i></p>
        <p><i>ient</i><i>yent</i></p>
        </présent>"""
    tense_elem: etree._Element = etree.fromstring(tense_elem_str)
    tense = Tense.fr.Présent
    tense_template = TenseTemplate(Lang.fr, mood, tense_elem)
    assert tense_template.mood == mood
    assert tense_template.name == str(tense.value)
    assert tense_template.person_endings[0].get_ending() == "ie"
    assert tense_template.person_endings[0].get_alternate_ending_if_available() == "ye"
    assert tense_template.person_endings[0].get_person() == "1s"
    assert tense_template.person_endings[3].get_ending() == "yons"
    assert (
        tense_template.person_endings[3].get_alternate_ending_if_available() == "yons"
    )
    assert tense_template.person_endings[3].get_person() == "1p"



from lxml import etree
from lxml.etree import Element

from verbecc.src.utils.string_utils import strip_accents
from verbecc.src.defs.types.exceptions import VerbsParserError


class Verb:
    def __init__(self, v_elem: etree._Element) -> None:
        if v_elem.tag != "v":
            raise VerbsParserError("Unexpected element")
        try:
            self.predicted = False
            self.pred_score = 1.0
            self.infinitive = ""
            e = v_elem.find("i", None)
            if e is not None:
                self.infinitive: str = e.text if e.text is not None else ""
            self.infinitive_no_accents = strip_accents(self.infinitive)
            self.template = ""
            e = v_elem.find("t", None)
            if e is not None:
                self.template: str = e.text if e.text is not None else ""
            self.translation_en = ""
            e = v_elem.find("en", None)
            if e is not None:
                self.translation_en: str = e.text if e.text is not None else ""
            self.impersonal = False
        except AttributeError as e:
            raise VerbsParserError(
                "Error parsing {}: {}".format(etree.tostring(v_elem), str(e))
            )

    def __repr__(self) -> str:
        return "infinitive={} infinitive_no_accents={} template={} translation_en={} impersonal={} predicted={} pred_score={}".format(
            self.infinitive,
            self.infinitive_no_accents,
            self.template,
            self.translation_en,
            self.impersonal,
            self.predicted,
            self.pred_score,
        )



import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


class LangCodeISO639_1(StrEnum):
    """
    Enum of supported languages as two-letter ISO 639-1 codes
    """

    ca = "ca"
    fr = "fr"
    es = "es"
    en = "en"
    it = "it"
    pt = "pt"
    ro = "ro"



import pytest
from lxml import etree
from typing import cast

from verbecc.src.defs.types.lang_code import LangCodeISO639_1 as Lang
from verbecc.src.defs.types.gender import Gender
from verbecc.src.defs.types.person import Person
from verbecc.src.conjugator.conjugator import Conjugator, AlternatesBehavior
from verbecc.src.parsers.tense_template import TenseTemplate
from verbecc.src.defs.types.exceptions import ConjugatorError
from verbecc.src.defs.types.conjugation import MoodsConjugation


@pytest.fixture(scope="module")
def cg():
    cg = Conjugator(lang="fr")
    yield cg


def test_all_verbs_have_templates(cg):
    verbs = cg.get_verbs()
    template_names = cg.get_template_names()
    missing_templates = set()
    for verb in verbs:
        if verb.template not in template_names:
            missing_templates.add(verb.template)
    assert len(missing_templates) == 0


def test_add_subjunctive_relative_prounoun(cg):
    assert (
        cg._inflector._add_subjunctive_relative_pronoun("tu manges", "")
        == "que tu manges"
    )
    assert (
        cg._inflector._add_subjunctive_relative_pronoun("il mange", "") == "qu'il mange"
    )
    assert (
        cg._inflector._add_subjunctive_relative_pronoun("elles mangent", "")
        == "qu'elles mangent"
    )


def test_add_reflexive_pronoun(cg):
    assert cg._inflector._add_reflexive_pronoun("lever") == "se lever"
    assert cg._inflector._add_reflexive_pronoun("écrouler") == "s'écrouler"


def test_split_reflexive(cg):
    assert cg._inflector._split_reflexive("se lever") == (True, "lever")
    assert cg._inflector._split_reflexive("s'écrouler") == (True, "écrouler")
    assert cg._inflector._split_reflexive("secouer") == (False, "secouer")


@pytest.mark.parametrize(
    "infinitive,expected_result",
    [
        ("être", False),
        ("lever", True),
        ("pleuvoir", False),
        ("manger", True),
    ],
)
def test_inflector_fr_verb_can_be_reflexive(cg, infinitive, expected_result):
    assert cg._inflector._verb_can_be_reflexive(infinitive) == expected_result


def test_inflector_fr_impersonal_verbs(cg):
    impersonal_verbs = [
        v.infinitive
        for v in cg._inflector._verb_parser.verbs
        if cg._inflector._is_impersonal_verb(v.infinitive)
    ]
    assert set(impersonal_verbs) == set(
        [
            "advenir",
            "apparoir",
            "bruiner",
            "bruire",
            "chaloir",
            "clore",
            "déclore",
            "échoir",
            "éclore",
            "enclore",
            "falloir",
            "forclore",
            "frire",
            "grêler",
            "messeoir",
            "neiger",
            "pleuvoir",
            "seoir",
            "sourdre",
        ]
    )


def test_inflector_fr_conjugate_simple_mood_tense(cg):
    mood = "indicatif"
    tense = "présent"
    verb_stem = "man"
    tense_elem = etree.fromstring(
        """<présent>
        <p><i>ge</i></p>
        <p><i>ges</i></p>
        <p><i>ge</i></p>
        <p><i>geons</i></p>
        <p><i>gez</i></p>
        <p><i>gent</i></p>
        </présent>""",
        parser=None,
    )
    tense_template = TenseTemplate(Lang.fr, mood, tense_elem)
    out = cg._conjugate_simple_mood_tense(verb_stem, mood, tense, tense_template)
    assert len(out) == 6
    assert out == [
        "je mange",
        "tu manges",
        "il mange",
        "nous mangeons",
        "vous mangez",
        "ils mangent",
    ]


def test_inflector_fr_get_verb_stem_from_template_name(cg):
    verb_stem = cg._inflector._get_verb_stem_from_template_name("manger", "man:ger")
    assert verb_stem == "man"
    verb_stem = cg._inflector._get_verb_stem_from_template_name("téléphoner", "aim:er")
    assert verb_stem == "téléphon"
    verb_stem = cg._inflector._get_verb_stem_from_template_name("vendre", "ten:dre")
    assert verb_stem == "ven"
    # In the case of irregular verbs, the verb stem is empty string
    verb_stem = cg._inflector._get_verb_stem_from_template_name("aller", ":aller")
    assert verb_stem == ""
    # The infinitive ending must match the template ending
    with pytest.raises(ConjugatorError):
        verb_stem = cg._inflector._get_verb_stem_from_template_name("vendre", "man:ger")


@pytest.mark.parametrize(
    "person,gender,is_reflexive,expected_result",
    [
        (Person.FirstPersonSingular, Gender.m, False, "je"),
        (Person.FirstPersonSingular, Gender.m, True, "je me"),
        (Person.SecondPersonSingular, Gender.m, False, "tu"),
        (Person.SecondPersonSingular, Gender.m, True, "tu te"),
        (Person.ThirdPersonSingular, Gender.m, False, "il"),
        (Person.ThirdPersonSingular, Gender.m, True, "il se"),
        (Person.ThirdPersonSingular, Gender.f, False, "elle"),
        (Person.ThirdPersonSingular, Gender.f, True, "elle se"),
        (Person.FirstPersonPlural, Gender.m, False, "nous"),
        (Person.FirstPersonPlural, Gender.m, True, "nous nous"),
        (Person.SecondPersonPlural, Gender.m, False, "vous"),
        (Person.SecondPersonPlural, Gender.m, True, "vous vous"),
        (Person.ThirdPersonPlural, Gender.m, False, "ils"),
        (Person.ThirdPersonPlural, Gender.m, True, "ils se"),
        (Person.ThirdPersonPlural, Gender.f, False, "elles"),
        (Person.ThirdPersonPlural, Gender.f, True, "elles se"),
    ],
)
def test_inflector_fr_get_default_pronoun(
    cg, person: Person, gender: Gender, is_reflexive: bool, expected_result: str
):
    assert (
        cg._inflector.get_default_pronoun(person, gender, is_reflexive=is_reflexive)
        == expected_result
    )


@pytest.mark.parametrize(
    "infinitive,expected_result",
    [
        (
            "avoir",
            ["j'ai", "tu as", "il a", "nous avons", "vous avez", "ils ont"],
        ),
        (
            "habiter",
            [
                "j'habite",
                "tu habites",
                "il habite",
                "nous habitons",
                "vous habitez",
                "ils habitent",
            ],
        ),
        (
            "s'habiller",
            [
                "je m'habille",
                "tu t'habilles",
                "il s'habille",
                "nous nous habillons",
                "vous vous habillez",
                "ils s'habillent",
            ],
        ),
    ],
)
def test_pronoun_combined_vowel_h_non_aspiré(cg, infinitive, expected_result):
    conj = cg.conjugate(infinitive)
    moods_conj = cast(MoodsConjugation, conj["moods"])
    assert moods_conj["indicatif"]["présent"] == expected_result


@pytest.mark.parametrize(
    "infinitive,expected_result",
    [
        (
            "habiter",
            [
                "que j'habite",
                "que tu habites",
                "qu'il habite",
                "que nous habitions",
                "que vous habitiez",
                "qu'ils habitent",
            ],
        )
    ],
)
def test_subjonctif_vowel_h_non_aspiré(cg, infinitive, expected_result):
    conj = cg.conjugate(infinitive)
    moods_conj = cast(MoodsConjugation, conj["moods"])
    assert moods_conj["subjonctif"]["présent"] == expected_result


def test_can_conjugate_all_verbs(cg):
    verbs = cg.get_verbs()
    all_conjugations = {}
    for verb in verbs:
        conjugation = cg.conjugate(verb.infinitive)
        all_conjugations[verb] = conjugation
    assert len(all_conjugations) == len(verbs)


def test_inflector_fr_conjugate_compound_raser(cg):
    infinitive = "raser"
    co = cg._get_conj_obs(infinitive)
    ret = cg._conjugate_compound(
        co,
        "subjonctif",
        "passé",
        "subjonctif",
        "présent",
        False,
        AlternatesBehavior.All,
        Gender.m,
        True,
    )
    assert ret == [
        ["que j'aie rasé"],
        ["que tu aies rasé"],
        ["qu'il ait rasé"],
        ["que nous ayons rasé"],
        ["que vous ayez rasé"],
        ["qu'ils aient rasé"],
    ]


def test_inflector_fr_conjugate_compound_se_raser(cg):
    """
    test targeting:
        - reflexive verb conjugation
        - compound verb conjugation with a verb conjugated with être (inflected participle)
        - Note: In French, all reflexive verbs are conjugated with être
    """
    infinitive = "se raser"
    co = cg._get_conj_obs(infinitive)
    ret = cg._conjugate_compound(
        co,
        "subjonctif",
        "passé",
        "subjonctif",
        "présent",
        False,
        AlternatesBehavior.All,
        Gender.m,
        True,
    )
    assert ret == [
        ["que je me sois rasé"],
        ["que tu te sois rasé"],
        ["qu'il se soit rasé"],
        ["que nous nous soyons rasés"],
        ["que vous vous soyez rasés"],
        ["qu'ils se soient rasés"],
    ]


def test_inflector_fr_conjugate_compound_parler_indicative_passé_composé(cg):
    """
    test targeting:
        - compound verb conjugation with a verb not conjugated with être (non-inflected participle)
    """
    infinitive = "parler"
    co = cg._get_conj_obs(infinitive)
    ret = cg._conjugate_compound(
        co,
        "indicatif",
        "passé-composé",
        "indicatif",
        "présent",
        False,
        AlternatesBehavior.All,
        Gender.m,
        True,
    )
    assert ret == [
        ["j'ai parlé"],
        ["tu as parlé"],
        ["il a parlé"],
        ["nous avons parlé"],
        ["vous avez parlé"],
        ["ils ont parlé"],
    ]



from typing import Dict, List, Tuple

from verbecc.src.defs.types.gender import Gender
from verbecc.src.defs.types.person import Person
from verbecc.src.defs.types.mood import MoodCa as Mood
from verbecc.src.defs.types.tense import TenseCa as Tense
from verbecc.src.defs.types import exceptions
from verbecc.src.defs.types.lang_code import LangCodeISO639_1
from verbecc.src.defs.types.lang_specific_options import (
    LangSpecificOptions,
)
from verbecc.src.inflectors import inflector
from verbecc.src.utils.string_utils import get_common_letter_count, strip_accents
from verbecc.src.conjugator.conjugation_object import ConjugationObjects


class InflectorCa(inflector.Inflector):
    @property
    def lang(self) -> LangCodeISO639_1:
        return LangCodeISO639_1.ca

    def __init__(self) -> None:
        super(InflectorCa, self).__init__()

    def _add_adverb_if_applicable(self, s: str, mood: Mood, tense: Tense) -> str:
        return s

    def get_default_pronoun(
        self,
        person: Person,
        gender: Gender = Gender.m,
        is_reflexive: bool = False,
        lang_specific_options: LangSpecificOptions = None,
    ) -> str:
        ret = ""
        if person == Person.FirstPersonSingular:
            ret = "jo"
            if is_reflexive:
                ret += " me"
        elif person == Person.SecondPersonSingular:
            ret = "tu"
            if is_reflexive:
                ret += " te"
        elif person == Person.ThirdPersonSingular:
            ret = "ell"
            if gender == Gender.f:
                ret = "ella"
            if is_reflexive:
                ret += " se"
        elif person == Person.FirstPersonPlural:
            ret = "nosaltres"
            if is_reflexive:
                ret += " nos"
        elif person == Person.SecondPersonPlural:
            ret = "vosaltres"
            if is_reflexive:
                ret += " os"
        elif person == Person.ThirdPersonPlural:
            ret = "ells"
            if gender == Gender.f:
                ret = "elles"
            if is_reflexive:
                ret += " se"
        return ret

    def _get_tenses_conjugated_without_pronouns(self) -> List[Tense]:
        return [
            Tense.Particip,
            Tense.Gerundi,
            Tense.InfinitiuPresent,
            Tense.ImperatiuPresent,
        ]

    def _get_auxilary_verb(
        self,
        co: ConjugationObjects,
        mood: Mood,
        tense: Tense,
    ) -> str:
        return "haver"

    def _get_infinitive_mood(self) -> Mood:
        return Mood.Infinitiu

    def _get_indicative_mood(self) -> Mood:
        return Mood.Indicatiu

    def _get_subjunctive_mood(self) -> Mood:
        return Mood.Subjuntiu

    def _get_conditional_mood(self) -> Mood:
        return Mood.Condicional

    def _get_participle_mood(self) -> Mood:
        return Mood.Participi

    def _get_participle_tense(self) -> Tense:
        return Tense.Particip

    def _get_alternate_hv_inflection(self, s: str) -> str:
        # if s.endswith('hay'):
        #     return s[:-1]
        return s

    def _get_compound_conjugations_aux_verb_map(
        self,
    ) -> Dict[str, Dict[str, Tuple[str, ...]]]:
        """
        TODO: Implement all these compound tenses (Spanish compound tenses in this commment, for reference)
        return {
            'indicatiu': {
                'pretèrit-perfet-compuest': ('indicatiu', 'present'),
                'pretèrit-pluscuamperfet': ('indicatiu', 'imperfet'),
                'pretèrit-anterior': ('indicatiu', 'pretèrit'),
                'futur-perfet': ('indicatiu', 'futuro')
            },
            'condicional': {
                'perfet': ('condicional', 'present')
            },
            'subjuntiu': {
                'pretèrit-perfet': ('subjuntiu', 'present'),
                'pretèrit-pluscuamperfet': ('subjuntiu', 'imperfet'),
                'futur-perfet': ('subjuntiu', 'futur')
            }
        }
        """
        return {}

    def _get_verb_stem_from_template_name(
        self, infinitive: str, template_name: str
    ) -> str:
        """Get the verb stem given an ininitive and a colon-delimited template name.
        E.g. infinitive='parlar' template_name='cant:ar' -> 'parl'

        Note: Base class _get_verb_stem_from_template_name raises exception if template
        ending doesn't match infinitive ending exactly but for Catalan, some verbs
        have endings where at least the first letter doesn't match.

        E.g. both 'jaure' and and 'jeure' are apparently conjugated
        identically, so we want either one to use the 'j:aure' template.
        So since this is Catalan, let it pass if the last n-1 letters of the
        template ending match the infinitive ending

        New problem: Template comen:çar ending doesn't match infinitive tòrcer

        Solution we'll just verify that, ignoring accents, the template either
        matches exactly or has at least len(template_ending)-1 characters in
        common.
        Ignoring accents, "çar" and "cer" have 2 characters in common which is at least 3-1
        "aure" and "eure" have 3 characters in common which is at least 4-1
        """
        _, template_ending = template_name.split(":")
        infinitive_no_accents = strip_accents(infinitive)
        template_ending_no_accents = strip_accents(infinitive)
        infinitive_ending_no_accents = infinitive_no_accents[-len(template_ending) :]
        if (
            not infinitive_ending_no_accents == template_ending_no_accents
            and not infinitive_no_accents[1:] == template_ending_no_accents[1:]
            and get_common_letter_count(
                infinitive_ending_no_accents, template_ending_no_accents
            )
            < len(template_ending) - 1
        ):
            raise exceptions.ConjugatorError(
                "Template '{}' ending doesn't "
                "match infinitive '{}', "
                "not even a little bit".format(template_name, infinitive)
            )
        return infinitive[: len(infinitive) - len(template_ending)]



import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


class VoseoOptions(StrEnum):
    """
    Enumeración de diferentes tipos de conjugación verbal en voseo en español

    NoVoseo: Indica que no se utiliza el voseo en la conjugación verbal.

    TIPO 1 Voseo ortodoxo, corresponde con la conjugación del voseo reverencial
    antiguo, aunque sin sus formas pronominales.
    Se escucha en Venezuela (Zulia), Colombia (occidente de Norte de Santander,
    sur de La Guajira y norte del Cesar, limítrofes con el Zulia venezolano),
    en el noroeste de Bolivia, en el centro de Panamá (Península de Azuero),
    y en una pequeña franja al oriente de Cuba.

    TIPO 2 Se escucha en la sierra de Ecuador, la zona meridional del Perú,
    noroeste de Argentina y el suroeste de Bolivia.

    TIPO 3 Ocurre en el sureste de México, en Centroamérica (excepto Panamá),
    la costa pacífica y zona andina centrooccidental de Colombia, parte de la
    zona andina de Venezuela, la costa de Ecuador, el sur y el este de Bolivia,
    Paraguay, Argentina y Uruguay. Es aceptado como parte de la norma culta en
    Argentina, Paraguay y Uruguay.

    TIPO 4 Ocurre en Argentina (Santiago del Estero).

    * En Chile nunca se pronuncia ni se escribe la «s» en las formas terminadas
    en «-áis», por ejemplo: «estái», «comái», «hablabai», etc. Existe la
    tendencia a utilizar la terminación en «-ay» para evitar la acentuación
    gráfica correspondiente a la terminación «-ái»; sin embargo, y pese a que
    la grafía con «-ay» se adapta mejor a la ortografía del español, no se
    recomienda para las formas diptongadas propias del voseo chileno.

    Desinencias en Indicativo Presente
    Tipo (nombres alternativos) 	-ar 	-er 	-ir
    --------------------------------------------------------
    Voseo tipo 1 (Voseo clásico) 	-áis 	-éis 	-ís
    Voseo tipo 2 	                -áis 	-ís 	-ís
    Voseo tipo 3 (Voseo típico) 	-ás 	-és 	-ís
    Voseo tipo 4 	                -ás 	-és 	-és
    Voseo chileno 	                -ái* 	-ís 	-ís

    Fuente: https://es.wikipedia.org/wiki/Voseo
    """

    NoVoseo = "no-voseo"  # default (tú)
    VoseoTipo1 = "voseo-tipo-1"  # Voseo Clásico
    VoseoTipo2 = "voseo-tipo-2"
    VoseoTipo3 = "voseo-tipo-3"  # Voseo típico aka Rioplatense
    VoseoTipo4 = "voseo-tipo-4"
    VoseoChileno = "voseo-chileno"



from __future__ import print_function
from bisect import bisect_left

try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree
from importlib_resources import as_file, files

# import gzip
import os

# import tempfile
from typing import List

from verbecc.src.parsers.verb import Verb
from verbecc.src.defs.types.exceptions import VerbNotFoundError, VerbsParserError
from verbecc.src.defs.types.lang_code import LangCodeISO639_1
from verbecc.src.defs.constants import config
from verbecc.src.mlconjug import mlconjug
from verbecc.src.utils import string_utils


class VerbsParser:
    def __init__(self, lang: LangCodeISO639_1 = LangCodeISO639_1.fr) -> None:
        self.verbs: List[Verb] = []
        parser = etree.XMLParser(encoding="utf-8", remove_blank_text=True, remove_comments=True)  # type: ignore
        source = files("verbecc.data.xml.verbs").joinpath("verbs-{}.xml".format(lang))
        with as_file(source) as fp:
            """
            with gzip.open(fp, "rt") as zf:
                with tempfile.NamedTemporaryFile(
                    prefix=f"/tmp/verbs-{lang}.xml.out.",
                    suffix=".xml",
                    mode="wt+",
                    encoding="utf-8",
                    delete=True,
                ) as tf:
                    next(zf)  # Skips the first line (gzip header plus xml header)
                    # Regenerate xml header
                    tf.write('<?xml version="1.0" encoding="utf-8"?>' + os.linesep)
                    for line in zf:
                        # there are some null bytes at the end that must be stripped
                        for byte in line:
                            if not byte.endswith("\x00"):
                                tf.write(byte)
                    tf.flush()
                    tree = etree.parse(tf.name, parser)  # type: ignore
            """
            tree = etree.parse(fp, parser)  # type: ignore
            root = tree.getroot()
            root_tag = "verbs-{}".format(lang)
            if root.tag != root_tag:
                raise VerbsParserError("Root XML Tag {} Not Found".format(root_tag))
            for child in root:
                if child.tag == "v":
                    self.verbs.append(Verb(child))  # type: ignore

            self.verbs = sorted(self.verbs, key=lambda v: v.infinitive)
            self._infinitives = [v.infinitive for v in self.verbs]
            self._verbs_no_accents = sorted(
                self.verbs, key=lambda v: v.infinitive_no_accents
            )
            self._infinitives_no_accents = [
                v.infinitive_no_accents for v in self._verbs_no_accents
            ]
            if config.ml:
                self.template_predictor = mlconjug.TemplatePredictor(
                    [(v.infinitive, v.template) for v in self.verbs], lang
                )

    def find_verb_by_infinitive(self, infinitive: str) -> Verb:
        """First try to find with accents, e.g. if infinitive is 'Abañar',
        search for 'abañar' and not 'abanar'.
        If not found then try searching with accents stripped.
        If all else fails, use machine-learning magic to predict
        which conjugation template should be used.
        """
        query = infinitive.lower()
        i = bisect_left(self._infinitives, query)
        if i != len(self._infinitives) and self._infinitives[i] == query:
            return self.verbs[i]
        query = string_utils.strip_accents(infinitive.lower())
        i = bisect_left(self._infinitives_no_accents, query)
        if (
            i != len(self._infinitives_no_accents)
            and self._infinitives_no_accents[i] == query
        ):
            return self._verbs_no_accents[i]
        if config.ml:
            template, pred_score = self.template_predictor.predict(query)
            verb_xml = "<v><i>{}</i><t>{}</t></v>".format(infinitive.lower(), template)
            ret = Verb(etree.fromstring(verb_xml))  # type: ignore
            ret.predicted = True
            ret.pred_score = pred_score
            return ret
        else:
            raise VerbNotFoundError

    def get_verbs_that_start_with(self, pre: str, max_results: int = 10) -> List[str]:
        ret: List[str] = []
        pre_no_accents = string_utils.strip_accents(pre.lower())
        for verb in self.verbs:
            if verb.infinitive_no_accents.startswith(pre_no_accents):
                ret.append(verb.infinitive)
                if len(ret) >= max_results:
                    break
        return ret



from typing import Dict, List, Tuple

from verbecc.src.conjugator.conjugation_object import ConjugationObjects
from verbecc.src.defs.types.gender import Gender
from verbecc.src.defs.types.lang_code import LangCodeISO639_1
from verbecc.src.defs.types.mood import MoodFr as Mood
from verbecc.src.defs.types.person import Person
from verbecc.src.defs.types.tense import TenseFr as Tense
from verbecc.src.defs.types.lang_specific_options import (
    LangSpecificOptions,
)
from verbecc.src.inflectors.inflector import Inflector
from verbecc.src.utils import string_utils

"""
DR & MRS VANDERTRAMPP verbs
"""
VERBS_CONJUGATED_WITH_ETRE = [
    "aller",
    "arriver",
    "descendre",
    "redescendre",
    "entrer",
    "rentrer",
    "monter",
    "remonter",
    "mourir",
    "naître",
    "renaître",
    "partir",
    "repartir",
    "passer",
    "rester",
    "retourner",
    "sortir",
    "ressortir",
    "tomber",
    "retomber",
    "venir",
    "devenir",
    "parvenir",
    "revenir",
]

VERBS_THAT_CANNOT_BE_REFLEXIVE_OTHER_THAN_IMPERSONAL_VERBS = ["être", "aller", "avoir"]


class InflectorFr(Inflector):
    @property
    def lang(self) -> LangCodeISO639_1:
        return LangCodeISO639_1.fr

    def __init__(self) -> None:
        super(InflectorFr, self).__init__()

    def get_verbs_that_start_with(self, query: str, max_results: int) -> List[str]:
        query = query.lower()
        is_reflexive, query = self._split_reflexive(query)
        matches = self._verb_parser.get_verbs_that_start_with(query, max_results)
        if is_reflexive:
            matches = [
                self._add_reflexive_pronoun(m)
                for m in matches
                if self._verb_can_be_reflexive(m)
            ]
        return matches

    def _is_impersonal_verb(self, infinitive: str) -> bool:
        ret = False
        verb = self.find_verb_by_infinitive(infinitive)
        template = self.find_template(verb.template)
        if (
            len(
                template.mood_templates[Mood.Indicatif]
                .tense_templates[Tense.Présent]
                .person_endings
            )
            < 6
        ):
            ret = True
        return ret

    def _verb_can_be_reflexive(self, infinitive: str) -> bool:
        return (
            not self._is_impersonal_verb(infinitive)
            and infinitive
            not in VERBS_THAT_CANNOT_BE_REFLEXIVE_OTHER_THAN_IMPERSONAL_VERBS
        )

    def _split_reflexive(self, infinitive: str) -> Tuple[bool, str]:
        """
        "se raser" => (True, "raser")
        "s'habiller" => (True, "habiller")
        "parler" => (False, "parler")
        """
        is_reflexive = False
        if infinitive.startswith("se "):
            is_reflexive = True
            infinitive = infinitive[3:]
        elif infinitive.startswith("s'"):
            is_reflexive = True
            infinitive = infinitive[2:]
        return is_reflexive, infinitive

    def _add_reflexive_pronoun(self, s: str) -> str:
        if string_utils.starts_with_vowel(s, h_is_vowel=True):
            return "s'" + s
        else:
            return "se " + s

    def _add_subjunctive_relative_pronoun(self, s: str, tense: Tense) -> str:
        if string_utils.starts_with_vowel(s, h_is_vowel=True):
            return "qu'" + s
        else:
            return "que " + s

    def _get_pronoun_suffix(self, person: Person, gender: Gender = Gender.m) -> str:
        return "-" + self.get_default_pronoun(person, gender).replace("tu", "toi")

    def get_default_pronoun(
        self,
        person: Person,
        gender: Gender = Gender.m,
        is_reflexive: bool = False,
        lang_specific_options: LangSpecificOptions = None,
    ) -> str:
        ret = ""
        if person == Person.FirstPersonSingular:
            ret = "je"
            if is_reflexive:
                ret += " me"
        elif person == Person.SecondPersonSingular:
            ret = "tu"
            if is_reflexive:
                ret += " te"
        elif person == Person.ThirdPersonSingular:
            ret = "il"
            if gender == Gender.f:
                ret = "elle"
            if is_reflexive:
                ret += " se"
        elif person == Person.FirstPersonPlural:
            ret = "nous"
            if is_reflexive:
                ret += " nous"
        elif person == Person.SecondPersonPlural:
            ret = "vous"
            if is_reflexive:
                ret += " vous"
        elif person == Person.ThirdPersonPlural:
            ret = "ils"
            if gender == Gender.f:
                ret = "elles"
            if is_reflexive:
                ret += " se"
        return ret

    def _get_tenses_conjugated_without_pronouns(self) -> List[Tense]:
        return [
            Tense.InfinitifPrésent,
            Tense.ParticipePresent,
            Tense.ImperatifPrésent,
            Tense.ParticipePassé,
        ]

    def _get_auxilary_verb(
        self,
        co: ConjugationObjects,
        mood: Mood,
        tense: Tense,
    ) -> str:
        ret = "avoir"
        if co.verb.infinitive in VERBS_CONJUGATED_WITH_ETRE or co.is_reflexive:
            ret = "être"
        return ret

    def _is_auxilary_verb_inflected(self, auxilary_verb: str) -> bool:
        return auxilary_verb == "être"

    def _get_infinitive_mood(self) -> Mood:
        return Mood.Infinitif

    def _get_indicative_mood(self) -> Mood:
        return Mood.Indicatif

    def _get_subjunctive_mood(self) -> Mood:
        return Mood.Subjonctif

    def _get_conditional_mood(self) -> Mood:
        return Mood.Conditionnel

    def _get_participle_mood(self) -> Mood:
        return Mood.Participe

    def _get_participle_tense(self) -> Tense:
        return Tense.ParticipePassé

    def _combine_pronoun_and_conj(self, pronoun: str, conj: str) -> str:
        ret = ""
        if pronoun[-1] == "e" and string_utils.starts_with_vowel(conj, h_is_vowel=True):
            ret += pronoun[:-1] + "'"
        else:
            ret += pronoun + " "
        ret += conj
        return ret

    def _add_present_participle_if_applicable(
        self, s: str, is_reflexive: bool, tense: Tense
    ) -> str:
        ret = s
        if is_reflexive and tense == self._get_participle_tense():
            ret += "étant "
        return ret

    def _add_reflexive_pronoun_or_pronoun_suffix_if_applicable(
        self,
        s: str,
        is_reflexive: bool,
        mood: Mood,
        tense: Tense,
        person: Person,
    ) -> str:
        if is_reflexive:
            if mood != Mood.Imperatif:
                s = self._add_reflexive_pronoun(s)
            else:
                s += self._get_pronoun_suffix(person)
        return s

    def _compound_conjugation_not_applicable(
        self, is_reflexive: bool, mood: Mood, hv_tense_name: Tense
    ) -> bool:
        return (
            is_reflexive
            and mood == Mood.Imperatif
            and hv_tense_name == Tense.ImperatifPrésent
        )

    def _get_compound_conjugations_aux_verb_map(
        self,
    ) -> Dict[Mood, Dict[Tense, Tuple[Mood, Tense]]]:
        """
        compound conjugations are formed using an auxiliary
        verb (aka helping verb)
        this method returns a Dictionary mapping of
        [compound-mood][compound-tense] to (aux-verb-mood, aux-verb-tense)
        """
        return {
            Mood.Indicatif: {
                Tense.PasséCompose: (Mood.Indicatif, Tense.Présent),
                Tense.PlusQueParfait: (Mood.Indicatif, Tense.Imparfait),
                Tense.FutureAntériuer: (Mood.Indicatif, Tense.FuturSimple),
                Tense.PasséAntérieur: (Mood.Indicatif, Tense.PasséSimple),
            },
            Mood.Subjonctif: {
                Tense.Passé: (Mood.Subjonctif, Tense.Présent),
                Tense.PlusQueParfait: (Mood.Subjonctif, Tense.Imparfait),
            },
            Mood.Conditionnel: {Tense.Passé: (Mood.Conditionnel, Tense.Présent)},
            Mood.Imperatif: {
                Tense.ImperatifPassé: (Mood.Imperatif, Tense.ImperatifPrésent)
            },
        }



import pytest
from typing import cast

from verbecc.src.conjugator.conjugator import Conjugator, MoodsConjugation
from verbecc.src.defs.types.alternates_behavior import AlternatesBehavior
from verbecc.src.defs.types.gender import Gender
from verbecc.src.defs.types.person import Person


@pytest.fixture(scope="module")
def cg():
    cg = Conjugator(lang="it")
    yield cg


def test_all_verbs_have_templates(cg):
    verbs = cg.get_verbs()
    template_names = cg.get_template_names()
    missing_templates = set()
    for verb in verbs:
        if verb.template not in template_names:
            missing_templates.add(verb.template)
    assert len(missing_templates) == 0


@pytest.mark.parametrize(
    "infinitive,mood,tense,expected_result",
    [
        (
            "avere",
            "indicativo",
            "presente",
            ["io ho", "tu hai", "lui ha", "noi abbiamo", "voi avete", "loro hanno"],
        ),
        (
            "avere",
            "indicativo",
            "imperfetto",
            [
                "io avevo",
                "tu avevi",
                "lui aveva",
                "noi avevamo",
                "voi avevate",
                "loro avevano",
            ],
        ),
        (
            "avere",
            "indicativo",
            "passato-remoto",
            [
                "io ebbi",
                "tu avesti",
                "lui ebbe",
                "noi avemmo",
                "voi aveste",
                "loro ebbero",
            ],
        ),
        (
            "avere",
            "indicativo",
            "futuro",
            [
                "io avrò",
                "tu avrai",
                "lui avrà",
                "noi avremo",
                "voi avrete",
                "loro avranno",
            ],
        ),
    ],
)
def test_inflector_it_conjugate_mood_tense(
    cg, infinitive, mood, tense, expected_result
):
    assert cg.conjugate_mood_tense(infinitive, mood, tense) == expected_result


def test_inflector_it_conjugate(cg):
    assert cg.conjugate("avere") != None


def test_inflector_it_add_subjunctive_relative_pronoun(cg):
    assert (
        cg._inflector._add_subjunctive_relative_pronoun("io abbia", "")
        == "che io abbia"
    )


@pytest.mark.parametrize(
    "person,gender,is_reflexive,expected_result",
    [
        (Person.FirstPersonSingular, Gender.m, False, "io"),
        (Person.FirstPersonSingular, Gender.m, True, "io mi"),
        (Person.SecondPersonSingular, Gender.m, False, "tu"),
        (Person.SecondPersonSingular, Gender.m, True, "tu ti"),
        (Person.ThirdPersonSingular, Gender.m, False, "lui"),
        (Person.ThirdPersonSingular, Gender.m, True, "lui si"),
        (Person.ThirdPersonSingular, Gender.f, False, "lei"),
        (Person.ThirdPersonSingular, Gender.f, True, "lei si"),
        (Person.FirstPersonPlural, Gender.m, False, "noi"),
        (Person.FirstPersonPlural, Gender.m, True, "noi ci"),
        (Person.SecondPersonPlural, Gender.m, False, "voi"),
        (Person.SecondPersonPlural, Gender.m, True, "voi vi"),
        (Person.ThirdPersonPlural, Gender.m, False, "loro"),
        (Person.ThirdPersonPlural, Gender.m, True, "loro si"),
        (Person.ThirdPersonPlural, Gender.f, False, "loro"),
        (Person.ThirdPersonPlural, Gender.f, True, "loro si"),
    ],
)
def test_inflector_it_get_default_pronoun(
    cg, person: Person, gender: Gender, is_reflexive: bool, expected_result: str
):
    assert (
        cg._inflector.get_default_pronoun(person, gender, is_reflexive=is_reflexive)
        == expected_result
    )


@pytest.mark.parametrize(
    "infinitive,expected_result",
    [
        (
            "avere",
            ["io ho", "tu hai", "lui ha", "noi abbiamo", "voi avete", "loro hanno"],
        ),
        (
            "essere",
            ["io sono", "tu sei", "lui è", "noi siamo", "voi siete", "loro sono"],
        ),
        (
            "alzare",
            [
                "io alzo",
                "tu alzi",
                "lui alza",
                "noi alziamo",
                "voi alzate",
                "loro alzano",
            ],
        ),
    ],
)
def test_indicative_present(cg, infinitive, expected_result):
    conj = cg.conjugate(infinitive)
    moods_conj = cast(MoodsConjugation, conj["moods"])
    assert moods_conj["indicativo"]["presente"] == expected_result


@pytest.mark.parametrize(
    "infinitive,expected_result",
    [
        (
            "avere",
            [
                "io ho avuto",
                "tu hai avuto",
                "lui ha avuto",
                "noi abbiamo avuto",
                "voi avete avuto",
                "loro hanno avuto",
            ],
        ),
        (
            "essere",
            [
                "io sono stato",
                "tu sei stato",
                "lui è stato",
                "noi siamo stati",
                "voi siete stati",
                "loro sono stati",
            ],
        ),
        (
            "alzare",
            [
                "io ho alzato",
                "tu hai alzato",
                "lui ha alzato",
                "noi abbiamo alzato",
                "voi avete alzato",
                "loro hanno alzato",
            ],
        ),
    ],
)
def test_passato_prossimo(cg, infinitive, expected_result):
    conj = cg.conjugate(infinitive)
    moods_conj = cast(MoodsConjugation, conj["moods"])
    assert moods_conj["indicativo"]["passato-prossimo"] == expected_result


@pytest.mark.parametrize(
    "infinitive,expected_result",
    [
        (
            "alzarsi",
            [
                "io mi alzo",
                "tu ti alzi",
                "lei si alza",
                "noi ci alziamo",
                "voi vi alzate",
                "loro si alzano",
            ],
        ),
    ],
)
def test_alzarsi_indicative_present(cg, infinitive, expected_result):
    conj = cg.conjugate(infinitive, gender=Gender.f)
    moods_conj = cast(MoodsConjugation, conj["moods"])
    assert moods_conj["indicativo"]["presente"] == expected_result


@pytest.mark.parametrize(
    "infinitive,expected_result",
    [
        (
            "alzarsi",
            [
                "io mi sono alzata",
                "tu ti sei alzata",
                "lei si è alzata",
                "noi ci siamo alzate",
                "voi vi siete alzate",
                "loro si sono alzate",
            ],
        ),
    ],
)
def test_inflector_it_alzarsi_indicativo_passato_prossimo(
    cg, infinitive, expected_result
):
    conj = cg.conjugate(infinitive, gender=Gender.f)
    moods_conj = cast(MoodsConjugation, conj["moods"])
    assert moods_conj["indicativo"]["passato-prossimo"] == expected_result


def test_inflector_it_conjugate_compound_essere_indicativo_passato_prossimo(cg):
    infinitive = "essere"
    co = cg._get_conj_obs(infinitive)
    ret = cg._conjugate_compound(
        co,
        "indicativo",
        "passato-prossimo",
        "indicativo",
        "presente",
        False,
        AlternatesBehavior.All,
        Gender.m,
        True,
    )
    assert ret == [
        ["io sono stato"],
        ["tu sei stato"],
        ["lui è stato"],
        ["noi siamo stati"],
        ["voi siete stati"],
        ["loro sono stati"],
    ]



from typing import Dict, List, Tuple

from verbecc.src.conjugator.conjugation_object import ConjugationObjects
from verbecc.src.defs.types.gender import Gender
from verbecc.src.defs.types.lang_code import LangCodeISO639_1
from verbecc.src.defs.types.mood import MoodPt as Mood
from verbecc.src.defs.types.person import Person
from verbecc.src.defs.types.tense import TensePt as Tense
from verbecc.src.defs.types.lang_specific_options import (
    LangSpecificOptions,
)
from verbecc.src.inflectors.inflector import Inflector


class InflectorPt(Inflector):
    @property
    def lang(self) -> LangCodeISO639_1:
        return LangCodeISO639_1.pt

    def __init__(self) -> None:
        super(InflectorPt, self).__init__()

    def _add_subjunctive_relative_pronoun(self, s: str, tense: Tense) -> str:
        if tense == Tense.Presente:
            return "que " + s
        elif tense == Tense.PretéritoImperfeito:
            return "se " + s
        elif tense == Tense.Futuro:
            return "quando " + s
        return s

    def _add_adverb_if_applicable(self, s: str, mood: Mood, tense: Tense) -> str:
        if mood == Mood.Imperativo and tense == Tense.Negativo:
            return "não " + s
        elif mood == Mood.Infinitivo and tense == Tense.InfinitivoPessoalPresente:
            return "por " + s
        return s

    def _add_reflexive_pronoun_or_pronoun_suffix_if_applicable(
        self,
        s: str,
        is_reflexive: bool,
        mood: Mood,
        tense: Tense,
        person: Person,
    ) -> str:
        imperative: bool = mood == Mood.Imperativo
        if imperative or (
            mood == Mood.Infinitivo and tense == Tense.InfinitivoPessoalPresente
        ):
            s += " " + self._get_pronoun_suffix(person, imperative=imperative)
        return s

    def get_default_pronoun(
        self,
        person: Person,
        gender: Gender = Gender.m,
        is_reflexive: bool = False,
        lang_specific_options: LangSpecificOptions = None,
    ) -> str:
        ret = ""
        if person == Person.FirstPersonSingular:
            ret = "eu"
            if is_reflexive:
                ret += " me"
        elif person == Person.SecondPersonSingular:
            ret = "tu"
            if is_reflexive:
                ret += " te"
        elif person == Person.ThirdPersonSingular:
            ret = "ele"
            if gender == Gender.f:
                ret = "ela"
            if is_reflexive:
                ret += " se"
        elif person == Person.FirstPersonPlural:
            ret = "nós"
            if is_reflexive:
                ret += " nos"
        elif person == Person.SecondPersonPlural:
            ret = "vós"
            if is_reflexive:
                ret += " se"
        elif person == Person.ThirdPersonPlural:
            ret = "eles"
            if gender == Gender.f:
                ret = "elas"
            if is_reflexive:
                ret += " se"
        return ret

    def _get_pronoun_suffix(
        self, person: Person, gender: Gender = Gender.m, imperative: bool = True
    ) -> str:
        ret = ""
        if person == Person.FirstPersonSingular:
            ret = "eu"
        if person == Person.SecondPersonSingular:
            ret = "tu"
        elif person == Person.ThirdPersonSingular:
            ret = "você"
            if not imperative:
                ret = "ele"
        elif person == Person.FirstPersonPlural:
            ret = "nós"
        elif person == Person.SecondPersonPlural:
            ret = "vós"
        elif person == Person.ThirdPersonPlural:
            ret = "vocês"
            if not imperative:
                ret = "eles"
        return ret

    def _get_tenses_conjugated_without_pronouns(self) -> List[Tense]:
        return [
            Tense.Particípio,
            Tense.Infinitivo,
            Tense.InfinitivoPessoalPresente,
            Tense.InfinitivoPessoalComposto,
            Tense.Afirmativo,
            Tense.Negativo,
            Tense.Gerúndio,
        ]

    def _get_auxilary_verb(
        self,
        co: ConjugationObjects,
        mood: Mood,
        tense: Tense,
    ) -> str:
        return "ter"

    def _get_infinitive_mood(self) -> Mood:
        return Mood.Infinitivo

    def _get_indicative_mood(self) -> Mood:
        return Mood.Indicativo

    def _get_subjunctive_mood(self) -> Mood:
        return Mood.Subjuntivo

    def _get_conditional_mood(self) -> Mood:
        return Mood.Condicional

    def _get_participle_mood(self) -> Mood:
        return Mood.Particípio

    def _get_participle_tense(self) -> Tense:
        return Tense.Particípio

    def _get_compound_conjugations_aux_verb_map(
        self,
    ) -> Dict[Mood, Dict[Tense, Tuple[Mood, Tense]]]:
        return {
            Mood.Indicativo: {
                Tense.PretéritoPerfeitoComposto: (Mood.Indicativo, Tense.Presente),
                Tense.PretéritoMaisQuePerfeitoComposto: (
                    Mood.Indicativo,
                    Tense.PretéritoImperfeito,
                ),
                Tense.PretéritoMaisQuePerfeitoAnterior: (
                    Mood.Indicativo,
                    Tense.PretéritoMaisQuePerfeito,
                ),
                Tense.FuturoDoPresenteComposto: (
                    Mood.Indicativo,
                    Tense.FuturoDoPresente,
                ),
            },
            Mood.Subjuntivo: {
                Tense.PretéritoPerfeito: (Mood.Subjuntivo, Tense.Presente),
                Tense.PretéritoMaisQuePerfeito: (
                    Mood.Subjuntivo,
                    Tense.PretéritoImperfeito,
                ),
                Tense.FuturoComposto: (Mood.Subjuntivo, Tense.Futuro),
            },
            Mood.Condicional: {
                Tense.FuturoDoPretéritoComposto: (
                    Mood.Condicional,
                    Tense.FuturoDoPretérito,
                )
            },
            Mood.Infinitivo: {
                Tense.InfinitivoPessoalComposto: (
                    Mood.Infinitivo,
                    Tense.InfinitivoPessoalPresente,
                )
            },
        }



"""
Brett Tolbert note:

This module adds an ML-based conjugation template prediction
feature. E.g. given the infinitive form of a verb, it can accurately
predict which conjugation template the verb should be conjugated
with.

The code in this module is based on an early version of mlconjug by Sekou Diao:
https://github.com/SekouD/mlconjug
Sekou Diao notes that a newer version of mjconjug is now available (mlconjug3):
https://github.com/SekouDiaoNlp/mlconjug3

However I have updated this version to use importlib_resources instead of
the deprecated pkg_resources API (as of today, mlconjug3 is still using
pkg_resources)

History:
verbecc predates mlconjug and verbecc's verb conjugation implementation
was developed independently of mlconjug, but credit to Sekou Diao for the ML
template prediction code in this module and for and the XML conjugation templates
for languages other than French and Catalan.

Credit to Pierre Sarrazin (Verbiste) for the developing the original French
XML conjugation template format on which both verbecc and mlconjug are based.

I found mlconjug and was impressed by the machine learning feature and I
so I borrowed this feature and retrofit it onto verbecc.
I chose not to add the entire mlconjug python package as a dependency because
it duplicates much of the functionality of verbecc and would be redundant.
mlconjug and verbecc are independent projects and this file, based on the
origin mlconjug module, has diverged.

verbecc is Open Source Software (GNU LGPL license)
mlconjug is also Open Source Software (MIT license)
Verbiste is Open Source Software (GNU GPL license)

Copyright (c) 2025, Brett Tolbert <http://bretttolbert.com/>
Copyright (c) 2017, SekouD <https://github.com/SekouDiaoNlp/>
Copyright (c) 2003-2016, Pierre Sarrazin <http://sarrazip.com/>
"""

__author__ = ["Sekou Diao"]
__credits__ = ["Sekou Diao", "Pierre Sarrazin", "Brett Tolbert"]


import re
from collections import defaultdict
from functools import partial
from importlib_resources import as_file, files
import os
import pickle
import random
from typing import Any, Dict, List, Tuple, Union
from zipfile import ZipFile

from sklearn.feature_selection import SelectFromModel
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.svm import LinearSVC
from sklearn.linear_model import SGDClassifier
from sklearn.pipeline import Pipeline

from verbecc.src.defs.constants.grammar_defines import ALPHABET
from verbecc.src.defs.types.lang_code import LangCodeISO639_1
import logging

from verbecc.src.defs.constants.config import DEVEL_MODE

logging_level = logging.CRITICAL + 1  # effectively disables logging
if DEVEL_MODE:
    logging_level = logging.DEBUG

logging.basicConfig(
    level=logging_level,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    handlers=[logging.FileHandler("verbecc-mlconjug.log"), logging.StreamHandler()],
)

logger = logging.getLogger(__name__)

VerbTemplatePair = Tuple[str, str]  # (verb, template)
Vectorizor = Union[CountVectorizer, Any]
FeatureSelector = Union[SelectFromModel, Any]
Classifier = Union[SGDClassifier, Any]


class TemplatePredictor:
    def __init__(
        self, verb_template_pairs: List[VerbTemplatePair], lang: LangCodeISO639_1
    ) -> None:
        self.data_set = DataSet(verb_template_pairs)
        model = load_model(lang)
        if not model:
            model = Model(lang=lang)
            model.train(self.data_set.train_input, self.data_set.train_labels)
            save_model(model)
        self.model = model
        return

    def predict(self, verb: str) -> Tuple[str, float]:
        prediction = self.model.predict([verb])[0]
        prediction_score = self.model.pipeline.predict_proba([verb])[0][prediction]
        template = self.data_set.templates[prediction]
        return (template, prediction_score)


class Model:
    """
    | This class manages the scikit-learn pipeline.
    | The Pipeline includes a feature vectorizer, a feature selector and a classifier.
    | If any of the vectorizer, feature selector or classifier is not supplied at instance declaration,
     the __init__ method will provide good default values that get more than 92% prediction accuracy.

    :param vectorizer: scikit-learn Vectorizer.
    :param feature_selector: scikit-learn Classifier with a fit_transform() method
    :param classifier: scikit-learn Classifier with a predict() method
    :param language: language of the corpus of verbs to be analyzed.
    """

    def __init__(
        self,
        vectorizer: Vectorizor = None,
        feature_selector: FeatureSelector = None,
        classifier: Classifier = None,
        lang: LangCodeISO639_1 = LangCodeISO639_1.fr,
    ) -> None:
        if not vectorizer:
            vectorizer = CountVectorizer(
                analyzer=partial(extract_verb_features, lang=lang, ngram_range=(2, 7)),
                binary=True,
            )
        if not feature_selector:
            feature_selector = SelectFromModel(
                LinearSVC(penalty="l1", max_iter=12000, dual=False, verbose=0)
            )
        if not classifier:
            classifier = SGDClassifier(
                loss="log_loss",
                penalty="elasticnet",
                l1_ratio=0.15,
                max_iter=40000,
                alpha=1e-5,
                verbose=0,
            )

        self.pipeline = Pipeline(
            [
                ("vectorizer", vectorizer),
                ("feature_selector", feature_selector),
                ("classifier", classifier),
            ]
        )
        self.lang = lang
        return

    def __repr__(self) -> str:
        return "{0}.{1}({2}, {3}, {4})".format(
            __name__, self.__class__.__name__, *sorted(self.pipeline.named_steps)
        )

    def train(self, samples: List[str], labels: List[int]) -> None:
        """
        Trains the pipeline on the supplied samples and labels.

        :param samples: list[str].
            List of verbs.
        :param labels: list[int].
            List of verb template indices.

        """
        self.pipeline = self.pipeline.fit(samples, labels)
        return

    def predict(self, verbs: List[str]) -> List[str]:
        """
        Predicts the conjugation class of the provided list of verbs.

        :param verbs: list[str].
            List of verbs.
        :return: list[str].
            List of predicted conjugation groups.

        """
        prediction = self.pipeline.predict(verbs)
        return prediction


class DataSet:
    """
    | This class holds and manages the data set.
    | Defines helper methodss for managing Machine Learning tasks like constructing a training and testing set.
    """

    def __init__(self, verb_template_pairs: List[VerbTemplatePair]) -> None:
        self.verbs = [pair[0] for pair in verb_template_pairs]
        self.templates = sorted(set([pair[1] for pair in verb_template_pairs]))
        self.dict_conjug = self._construct_dict_conjug(verb_template_pairs)
        self._split_test_train()
        return

    def _construct_dict_conjug(
        self, verb_template_pairs: List[VerbTemplatePair]
    ) -> Dict[str, List[str]]:
        """
        | Populates the dictionary containing the conjugation templates.
        | Populates the lists containing the verbs and their templates.

        :param verb_template_pairs: list.
            List of tuples of (verb,template) e.g. ('abaisser','aim:er')
        :return: defaultdict.
            defaultdict mapping each template to one or more verbs e.g. {'aim:er': ['abaisser', ...]}
        """
        ret = defaultdict(list)
        random.shuffle(verb_template_pairs)
        for verb, template in verb_template_pairs:
            ret[template].append(verb)
        return ret

    def _split_test_train(self, threshold: int = 8, proportion: float = 0.5) -> None:
        """
        Splits the template:verbs dict into a training and a testing set.

        :param verb_template_pairs: list.
            List of tuples of (verb,template) e.g. ('abaisser','aim:er')
        :param threshold: int.
            Minimum size of conjugation class to be split.
        :param proportion: float.
            Proportion of samples in the training set.
            Must be between 0 and 1.

        """
        if proportion <= 0 or proportion > 1:
            raise ValueError(
                f"The split proportion ({proportion}) must be between 0 and 1."
            )
        self.min_threshold = threshold
        self.split_proportion = proportion
        train_set: List[VerbTemplatePair] = []
        test_set: List[VerbTemplatePair] = []
        for template, lverbs in self.dict_conjug.items():
            if len(lverbs) <= threshold:
                for verb in lverbs:
                    train_set.append((verb, template))
            else:
                index = round(len(lverbs) * proportion)
                for verb in lverbs[:index]:
                    train_set.append((verb, template))
                for verb in lverbs[index:]:
                    test_set.append((verb, template))
        random.shuffle(train_set)
        random.shuffle(test_set)
        self.train_input: List[str] = [elmt[0] for elmt in train_set]
        self.train_labels: List[int] = [
            self.templates.index(elmt[1]) for elmt in train_set
        ]
        self.test_input: List[str] = [elmt[0] for elmt in test_set]
        self.test_labels: List[int] = [
            self.templates.index(elmt[1]) for elmt in test_set
        ]


def extract_verb_features(
    verb: str, lang: LangCodeISO639_1, ngram_range: Tuple[int, int]
) -> List[str]:
    """
    | Custom Vectorizer optimized for extracting verbs features.
    | The Vectorizer subclasses sklearn.feature_extraction.text.CountVectorizer .
    | As in Indo-European languages verbs are inflected by adding a morphological suffix,
     the vectorizer extracts verb endings and produces a vector representation of the verb with binary features.

    | To enhance the results of the feature extration, several other features have been included:

    | The features are the verb's ending n-grams, starting n-grams, length of the verb, number of vowels,
     number of consonants and the ratio of vowels over consonants.

    :param verb: string.
        Verb to vectorize.
    :param lang: LangCodeISO639_1.
        Language to analyze.
    :param ngram_range: tuple.
        The range of the ngram sliding window.
    :return: list[str].
        List of the most salient features of the verb for the task of finding it's conjugation's class.

    """
    _white_spaces = re.compile(r"\s\s+")
    verb = _white_spaces.sub(" ", verb)
    verb = verb.lower()
    verb_len = len(verb)
    length_feature = "LEN={0}".format(str(verb_len))
    min_n, max_n = ngram_range
    final_ngrams = [
        "END={0}".format(verb[-n:]) for n in range(min_n, min(max_n + 1, verb_len + 1))
    ]
    initial_ngrams = [
        "START={0}".format(verb[:n]) for n in range(min_n, min(max_n + 1, verb_len + 1))
    ]
    if lang not in ALPHABET:
        lang = "en"  # We chose 'en' as the default alphabet because EN is more standard, without accents or diactrics.
    vowels = sum(verb.count(c) for c in ALPHABET[lang]["vowels"])
    vowels_number = "VOW_NUM={0}".format(vowels)
    consonants = sum(verb.count(c) for c in ALPHABET[lang]["consonants"])
    consonants_number = "CONS_NUM={0}".format(consonants)
    if consonants == 0:
        vow_cons_ratio = "V/C=N/A"
    else:
        vow_cons_ratio = "V/C={0}".format(round(vowels / consonants, 2))
    final_ngrams.extend(initial_ngrams)
    final_ngrams.extend(
        (length_feature, vowels_number, consonants_number, vow_cons_ratio)
    )
    return final_ngrams


def get_model_zip_filename(lang: LangCodeISO639_1) -> str:
    return "data/models/trained_model-{}.zip".format(lang)


def get_model_pickle_filename(lang: LangCodeISO639_1) -> str:
    return "trained_model-{0}.pickle".format(lang)


def save_model(model: Model) -> None:
    pickle_filename = get_model_pickle_filename(model.lang)
    with open(pickle_filename, "wb") as f:
        pickle.dump(model, f)
    zip_filename = get_model_zip_filename(model.lang)
    zip_path = files("verbecc") / zip_filename
    with as_file(zip_path) as f:
        with ZipFile(f, mode="w") as zf:
            zf.write(pickle_filename)
            logger.info(
                "Saved model pickle filename %s to zip filename %s.",
                pickle_filename,
                zip_filename,
            )
    os.remove(pickle_filename)


def load_model(lang: LangCodeISO639_1) -> Model:
    model = None
    zip_filename = get_model_zip_filename(lang)
    try:
        zip_path = files("verbecc") / zip_filename
        with as_file(zip_path) as f:
            with ZipFile(f) as zf:
                pickle_filename = get_model_pickle_filename(lang)
                with zf.open(pickle_filename, "r") as model_pickle:
                    model = pickle.loads(model_pickle.read())
                    logger.info(
                        "Loaded model pickle filename %s from zip filename %s",
                        pickle_filename,
                        zip_filename,
                    )
    except Exception as ex:
        logger.warning(
            "Exception loading model %s: %s", zip_filename, ex, exc_info=True
        )
    return model



"""
This script was used to remove unwanted tense elements
(e.g. compound tenses) from the mlconjug conjugation XML files
"""

from lxml import etree, objectify
import os
from typing import List
from verbecc.src.defs.types.tense import Tense
from verbecc.src.defs.types.mood import Mood

working_dir = "../verbecc/data"
in_file = "conjugations-ro.xml"
out_file = "conjugations-ro.mod.xml"


def remove_mood_tense(root: etree._Element, tenses_to_remove: List[Tense]) -> None:
    removed_elem_cnt = 0
    for template_elem in root:
        if template_elem.tag == "template":
            for mood_elem in template_elem:
                for tense_elem in mood_elem:
                    if tense_elem.tag in tenses_to_remove:
                        mood_elem.remove(tense_elem)
                        removed_elem_cnt += 1
    print("removed {} elements".format(removed_elem_cnt))


def remove_mood(root: etree._Element, moods_to_remove: List[Mood]) -> None:
    removed_elem_cnt = 0
    for template_elem in root:
        if template_elem.tag == "template":
            for mood_elem in template_elem:
                if mood_elem.tag in moods_to_remove:
                    template_elem.remove(mood_elem)
                    removed_elem_cnt += 1
    print("removed {} elements".format(removed_elem_cnt))


def move_tense(
    root: etree._Element,
    tense: Tense,
    old_mood: Mood,
    new_mood: Mood,
    remove_old_mood: bool,
) -> None:
    moved_elem_cnt = 0
    for template_elem in root:
        if template_elem.tag == "template":
            tense_elem_to_move = None
            # find tense to move
            for mood_elem in template_elem:
                if mood_elem.tag == old_mood:
                    for tense_elem in mood_elem:
                        if tense_elem.tag == tense:
                            tense_elem_to_move = tense_elem
                            mood_elem.remove(tense_elem)
                            if remove_old_mood:
                                mood_elem.getparent().remove(mood_elem)
                            break
            # Now move it
            if tense_elem_to_move is not None:
                for mood_elem in template_elem:
                    if mood_elem.tag == new_mood:
                        mood_elem.append(tense_elem)
                        moved_elem_cnt += 1
    print("moved {} elements".format(moved_elem_cnt))


def main() -> None:
    parser = etree.XMLParser(dtd_validation=False, encoding="utf-8")
    tree = etree.parse(os.path.join(working_dir, in_file), parser)
    root = tree.getroot()

    remove_mood(root, ["Conditional"])
    # remove_mood_tense(root, ['perfect'])
    # move_tense(root, 'Viitor-II-popular', 'Viitor', 'Indicativ', True)

    with open(os.path.join(working_dir, out_file), "wb") as f:
        objectify.deannotate(root, cleanup_namespaces=True)
        xml = etree.tostring(
            tree,
            encoding="utf-8",
            method="xml",
            pretty_print=True,
            xml_declaration=True,
        )
        f.write(xml)


if __name__ == "__main__":
    main()



import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


class ParticipleInflection(StrEnum):
    MasculineSingular = "ms"
    MasculinePlural = "mp"
    FeminineSingular = "fs"
    FemininePlural = "fp"



import copy
from typing import cast, Dict, List, Tuple

from verbecc.src.defs.types.gender import Gender
from verbecc.src.defs.types.lang_code import LangCodeISO639_1
from verbecc.src.defs.types.person import Person
from verbecc.src.defs.types.mood import MoodEs as Mood
from verbecc.src.defs.types.tense import TenseEs as Tense
from verbecc.src.defs.types.lang_specific_options import (
    LangSpecificOptions,
)
from verbecc.src.defs.types.lang.es.lang_specific_options_es import (
    LangSpecificOptionsEs,
)
from verbecc.src.defs.types.lang.es.voseo_options import VoseoOptions
from verbecc.src.inflectors.inflector import Inflector
from verbecc.src.conjugator.conjugation_object import ConjugationObjects
from verbecc.src.parsers.tense_template import TenseTemplate
from verbecc.src.parsers.person_ending import PersonEnding


class InflectorEs(Inflector):
    @property
    def lang(self) -> LangCodeISO639_1:
        return LangCodeISO639_1.es

    def __init__(self) -> None:
        super(InflectorEs, self).__init__()

    def _add_adverb_if_applicable(self, s: str, mood: Mood, tense: Tense) -> str:
        if mood == Mood.Imperativo and tense == Tense.Negativo:
            return "no " + s
        return s

    def get_default_pronoun(
        self,
        person: Person,
        gender: Gender = Gender.m,
        is_reflexive: bool = False,
        lang_specific_options: LangSpecificOptions = None,
    ) -> str:
        lang_opts = None
        if lang_specific_options is not None:
            lang_opts = cast(LangSpecificOptionsEs, lang_specific_options)

        ret = ""
        if person == Person.FirstPersonSingular:
            ret = "yo"
            if is_reflexive:
                ret += " me"
        elif person == Person.SecondPersonSingular:
            if (
                lang_opts is not None
                and lang_opts.voseo_options != VoseoOptions.NoVoseo
            ):
                ret = "vos"
            else:
                ret = "tú"
            if is_reflexive:
                ret += " te"
        elif person == Person.ThirdPersonSingular:
            ret = "él"
            if gender == Gender.f:
                ret = "ella"
            if is_reflexive:
                ret += " se"
        elif person == Person.FirstPersonPlural:
            ret = "nosotros"
            if is_reflexive:
                ret += " nos"
        elif person == Person.SecondPersonPlural:
            ret = "vosotros"
            if is_reflexive:
                ret += " os"
        elif person == Person.ThirdPersonPlural:
            ret = "ellos"
            if gender == Gender.f:
                ret = "ellas"
            if is_reflexive:
                ret += " se"
        return ret

    def _get_tenses_conjugated_without_pronouns(self) -> List[Tense]:
        return [
            Tense.Participo,
            Tense.Gerundio,
            Tense.Infinitivo,
            Tense.Afirmativo,
            Tense.Negativo,
        ]

    def _get_auxilary_verb(
        self,
        co: ConjugationObjects,
        mood: Mood,
        tense: Tense,
    ) -> str:
        return "haber"

    def _get_infinitive_mood(self) -> Mood:
        return Mood.Infinitivo

    def _get_indicative_mood(self) -> Mood:
        return Mood.Indicativo

    def _get_subjunctive_mood(self) -> Mood:
        return Mood.Subjuntivo

    def _get_conditional_mood(self) -> Mood:
        return Mood.Condicional

    def _get_participle_mood(self) -> Mood:
        return Mood.Participo

    def _get_participle_tense(self) -> Mood:
        return Tense.Participo

    def _get_alternate_hv_inflection(self, s: str) -> str:
        if s.endswith("hay"):
            return s[:-1]
        return s

    def _get_compound_conjugations_aux_verb_map(
        self,
    ) -> Dict[Mood, Dict[Tense, Tuple[Mood, Tense]]]:
        return {
            Mood.Indicativo: {
                Tense.PretéritoPerfectoCompuesto: (
                    Mood.Indicativo,
                    Tense.Presente,
                ),
                Tense.PretéritoPluscuamperfecto: (
                    Mood.Indicativo,
                    Tense.PretéritoImperfecto,
                ),
                Tense.PretéritoAnterior: (
                    Mood.Indicativo,
                    Tense.PretéritoPerfectoSimple,
                ),
                Tense.FuturoPerfecto: (Mood.Indicativo, Tense.Futuro),
            },
            Mood.Condicional: {Tense.Perfecto: (Mood.Condicional, Tense.Presente)},
            Mood.Subjuntivo: {
                Tense.PretéritoPerfecto: (Mood.Subjuntivo, Tense.Presente),
                Tense.PretéritoPluscuamperfecto1: (
                    Mood.Subjuntivo,
                    Tense.PretéritoImperfecto1,
                ),
                Tense.PretéritoPluscuamperfecto2: (
                    Mood.Subjuntivo,
                    Tense.PretéritoImperfecto2,
                ),
                Tense.FuturoPerfecto: (Mood.Subjuntivo, Tense.Futuro),
            },
        }

    def modify_person_ending_if_applicable(
        self,
        person_ending: PersonEnding,
        mood: Mood,
        tense: Tense,
        tense_template: TenseTemplate,
        lang_specific_options: LangSpecificOptions,
    ) -> PersonEnding:
        """
        Hook for certain languages e.g. Spanish that modify
        the standard person endings for certain persons depending
        on language specific options (e.g. Voseo)

        Vos may be used in place of tú in the second person singular
        with corresponding changes to verb endings.
        The verb endings for vos are different from those for tú in
        the present indicative, present subjunctive, and imperative moods.
        Vos is conjugated like the vosotros form but without the 'i' in the ending.
        For example, the present indicative forms are:

        vosotros bebéis -> vos bebés
        vosotros habláis -> vos hablás
        vosotros dormís -> vos dormís

        "éis" -> "és"
        "áis" -> "ás"
        "ís" -> "ís"

        For the imperativo affirmativo we just drop the 'd' and accent the vowel e.g.
        (vosotros) hablad -> (vos) hablá
        The imperativo negativo is the same as tú.

        """
        # map of vosotros endings to vos endings
        # for the present indicative, present subjunctive
        VOSEO_ENDINGS_MAP_INDICATIVE_OR_SUBJUNCTIVE_PRESENT: Dict[str, str] = {
            "as": "ás",
            "es": "és",
            "ís": "ís",
            "áis": "ás",
            "éis": "és",
        }
        VOSEO_ENDINGS_MAP_IMPERATIVE: Dict[str, str] = {"ad": "á", "id": "í", "ed": "é"}
        lang_opts = None
        if lang_specific_options is not None:
            lang_opts = cast(LangSpecificOptionsEs, lang_specific_options)
            if (
                lang_opts is not None
                and lang_opts.voseo_options != VoseoOptions.NoVoseo
                and person_ending.person == Person.SecondPersonSingular
            ):
                if lang_opts.voseo_options != VoseoOptions.VoseoTipo3:
                    # only voseo tipo 3 (voseo típico aka Rioplatense) is supported at the moment
                    raise NotImplementedError

                if (
                    (mood == Mood.Indicativo and tense == Tense.Presente)
                    or (mood == Mood.Subjuntivo and tense == Tense.Presente)
                    or (mood == Mood.Imperativo and tense == Tense.Afirmativo)
                ):
                    # first replace with given SecondPersonSingular (tú) ending(s)
                    # with the SecondPersonPlural (vosotros) ending(s)
                    replacement_person_ending = copy.deepcopy(
                        tense_template.get_person_ending(Person.SecondPersonPlural)
                    )
                    # change replacement PersonEnding Person from second person plural to singular
                    replacement_person_ending.person = Person.SecondPersonSingular

                    # modify the endings for voseo
                    for i, ending in enumerate(replacement_person_ending.get_endings()):
                        endings_map = (
                            VOSEO_ENDINGS_MAP_INDICATIVE_OR_SUBJUNCTIVE_PRESENT
                        )
                        if mood == Mood.Imperativo:
                            endings_map = VOSEO_ENDINGS_MAP_IMPERATIVE
                        for e in endings_map:
                            # e.g. if ending is "sed", it endswith "ed", so "s" + "é" -> "sé"
                            if ending.endswith(e):
                                replacement_person_ending.endings[i] = (
                                    ending[: -len(e)] + endings_map[e]
                                )
                                break
                    return replacement_person_ending
        return person_ending



from verbecc.src.defs.types.exceptions import InvalidLangError
from verbecc.src.defs.types.lang_code import LangCodeISO639_1
from verbecc.src.inflectors.inflector import Inflector
from verbecc.src.inflectors.lang.inflector_ca import InflectorCa
from verbecc.src.inflectors.lang.inflector_es import InflectorEs
from verbecc.src.inflectors.lang.inflector_fr import InflectorFr
from verbecc.src.inflectors.lang.inflector_it import InflectorIt
from verbecc.src.inflectors.lang.inflector_pt import InflectorPt
from verbecc.src.inflectors.lang.inflector_ro import InflectorRo


class InflectorFactory:
    @classmethod
    def make_inflector(cls, lang: LangCodeISO639_1) -> Inflector:
        """
        :param lang: two-letter language code (ISO 639-1 Code)
        :type lang: LangCodeISO639_1
        """
        ret = None
        if lang == "ca":
            ret = InflectorCa()
        elif lang == "es":
            ret = InflectorEs()
        elif lang == "fr":
            ret = InflectorFr()
        elif lang == "it":
            ret = InflectorIt()
        elif lang == "pt":
            ret = InflectorPt()
        elif lang == "ro":
            ret = InflectorRo()
        else:
            raise InvalidLangError
        return ret



from collections import Counter
from typing import Dict
import unicodedata


def strip_accents(s: str) -> str:
    return "".join(
        c for c in unicodedata.normalize("NFD", s) if unicodedata.category(c) != "Mn"
    )


def starts_with_vowel(s: str, h_is_vowel: bool = False) -> bool:
    if len(s) == 0:
        return False
    vowels = ["a", "e", "i", "o", "u"]
    if h_is_vowel:
        vowels.append("h")
    return strip_accents(s)[0] in vowels


def get_common_letters(s1: str, s2: str) -> Dict[str, int]:
    common_letters = Counter(s1) & Counter(s2)  # => {'q': 2, 'r': 1}
    return common_letters


def get_common_letter_count(s1: str, s2: str) -> int:
    return sum(get_common_letters(s1, s2).values())


def unicodefix(s: str) -> str:
    # Fix Python 2.x.
    try:
        return s.decode("utf-8")
    except AttributeError:
        return s



from verbecc.src.utils.string_utils import (
    get_common_letters,
    get_common_letter_count,
    starts_with_vowel,
    strip_accents,
    unicodefix,
)


def test_starts_with_vowel():
    assert starts_with_vowel("aller")
    assert not starts_with_vowel("banane")
    assert starts_with_vowel("éparpiller")
    assert not starts_with_vowel("yodler")
    assert not starts_with_vowel("")


def test_strip_accents():
    assert (
        strip_accents("français, être, égalité, très")
        == "francais, etre, egalite, tres"
    )


def test_get_common_letters():
    assert get_common_letters("brett", "tolbert") == {"b": 1, "e": 1, "r": 1, "t": 2}


def test_get_common_letters_count():
    assert get_common_letter_count("brett", "tolbert") == 5
    assert get_common_letter_count("çar", "cer") == 1
    assert get_common_letter_count(strip_accents("çar"), "cer") == 2


def test_unicodefix():
    assert unicodefix("éparpiller") == "éparpiller"



from verbecc.src.parsers.verb import Verb
from verbecc.src.parsers.conjugation_template import ConjugationTemplate


class ConjugationObjects:
    def __init__(
        self,
        infinitive: str,
        verb: Verb,
        template: ConjugationTemplate,
        verb_stem: str,
        is_reflexive: bool,
    ) -> None:
        """
        :param verb_stem: the verb stem after applicable template
                            stem modifications i.e. modify-stem="strip-accents"
        :type verb_stem: str
        """
        self.infinitive = infinitive
        self.verb = verb
        self.template = template
        self.verb_stem = verb_stem
        self.is_reflexive = is_reflexive

    def __repr__(self) -> str:
        return "infinitive={} verb={} template={} verb_stem={} is_reflexive={}".format(
            self.infinitive,
            self.verb,
            self.template,
            self.verb_stem,
            self.is_reflexive,
        )



import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

from typing import Dict, List, Tuple

from verbecc.src.defs.types.lang_code import LangCodeISO639_1
from verbecc.src.defs.types.partiple_inflection import ParticipleInflection
from verbecc.src.defs.types.person import Person

# map of ISO 639-1 codes to long names (in target language)
SUPPORTED_LANGUAGES: Dict[LangCodeISO639_1, str] = {
    LangCodeISO639_1.ca: "català",
    LangCodeISO639_1.es: "español",
    LangCodeISO639_1.fr: "français",
    LangCodeISO639_1.it: "italiano",
    LangCodeISO639_1.pt: "português",
    LangCodeISO639_1.ro: "română",
}

# Order of grammatical persons in data structures
# Note: 'usted' is 3s despite being semantically 2p
PERSONS: Tuple[Person, Person, Person, Person, Person, Person] = (
    Person.FirstPersonSingular,
    Person.SecondPersonSingular,
    Person.ThirdPersonSingular,
    Person.FirstPersonPlural,
    Person.SecondPersonPlural,
    Person.ThirdPersonPlural,
)

# e.g. manger : mange, mangeons, mangez
IMPERATIVE_PERSONS_FR: Tuple[Person, Person, Person] = (
    Person.SecondPersonSingular,
    Person.FirstPersonPlural,
    Person.SecondPersonPlural,
)

# e.g. habla, hable, hablemos, hablad, hablen
IMPERATIVE_PERSONS_ES: Tuple[Person, Person, Person, Person, Person] = (
    Person.SecondPersonSingular,
    Person.ThirdPersonSingular,
    Person.FirstPersonPlural,
    Person.SecondPersonPlural,
    Person.ThirdPersonPlural,
)

# e.g. habla, habli, hablem, hableu, hablin
IMPERATIVE_PERSONS_CA = IMPERATIVE_PERSONS_ES

# e.g. -, sii, sia, siamo, siate, siano
# Italian currently has the "-" placeholder for the 1PS unlike French and Spanish
# TODO: Make them consistent
IMPERATIVE_PERSONS_IT: Tuple[Person, Person, Person, Person, Person, Person] = (
    Person.FirstPersonSingular,
    Person.SecondPersonSingular,
    Person.ThirdPersonSingular,
    Person.FirstPersonPlural,
    Person.SecondPersonPlural,
    Person.ThirdPersonPlural,
)

# e.g. -, sê tu, seja você, sejamos nós, sede vós, sejam vocês
# Portuguese currently has the "-" placeholder for the 1PS like Italian and unlike French and Spanish
# TODO: Make them consistent
IMPERATIVE_PERSONS_PT = IMPERATIVE_PERSONS_IT

# The Romanian imperative is used for commands and requests and
# is formed primarily using the second person verb forms.
IMPERATIVE_PERSONS_RO: Tuple[Person, Person] = (
    Person.SecondPersonSingular,
    Person.SecondPersonPlural,
)

IMPERATIVE_PERSONS: Dict[LangCodeISO639_1, List[Person]] = {
    LangCodeISO639_1.fr: IMPERATIVE_PERSONS_FR,
    LangCodeISO639_1.es: IMPERATIVE_PERSONS_ES,
    LangCodeISO639_1.ca: IMPERATIVE_PERSONS_CA,
    LangCodeISO639_1.it: IMPERATIVE_PERSONS_IT,
    LangCodeISO639_1.pt: IMPERATIVE_PERSONS_PT,
    LangCodeISO639_1.ro: IMPERATIVE_PERSONS_IT,
}

# Order of participle inflections in XML
PARTICIPLE_INFLECTIONS: Tuple[
    ParticipleInflection,
    ParticipleInflection,
    ParticipleInflection,
    ParticipleInflection,
] = (
    ParticipleInflection.MasculineSingular,
    ParticipleInflection.MasculinePlural,
    ParticipleInflection.FeminineSingular,
    ParticipleInflection.FemininePlural,
)

ALPHABET = {
    LangCodeISO639_1.fr: {
        "vowels": "aáàâeêéèiîïoôöœuûùy",
        "consonants": "bcçdfghjklmnpqrstvwxyz",
    },
    LangCodeISO639_1.en: {
        "vowels": "aeiouy",
        "consonants": "bcdfghjklmnpqrstvwxyz",
    },
    LangCodeISO639_1.ca: {
        "vowels": "aáàâeéèiïoôuûùy",
        "consonants": "bcdfghjklmnñpqrstvwxyz",
    },
    LangCodeISO639_1.es: {
        "vowels": "aáeiíoóuúy",
        "consonants": "bcdfghjklmnñpqrstvwxyz",
    },
    LangCodeISO639_1.es: {
        "vowels": "aàeéèiìîoóòuùy",
        "consonants": "bcdfghjklmnpqrstvwxyz",
    },
    LangCodeISO639_1.pt: {
        "vowels": "aàãááeêéiíoóõuúy",
        "consonants": "bcçdfghjklmnpqrstvwxyz",
    },
    LangCodeISO639_1.ro: {
        "vowels": "aăâeiîouy",
        "consonants": "bcdfghjklmnpqrsșştțţvwxyz",
    },
}



from lxml import etree
from typing import Dict

from verbecc.src.parsers.mood_template import MoodTemplate
from verbecc.src.defs.types.exceptions import ConjugationTemplateError
from verbecc.src.defs.types.mood import Mood
from verbecc.src.defs.types.lang_code import LangCodeISO639_1 as Lang


class ConjugationTemplate:
    def __init__(self, lang: Lang, template_elem: etree._Element) -> None:
        if template_elem.tag != "template":
            raise ConjugationTemplateError("Unexpected element")
        try:
            name_attrib = template_elem.get("name", default=None)
            self.name = str(name_attrib)
            self.mood_templates: Dict[Mood, MoodTemplate] = {}
            for mood_elem in template_elem:  # type: ignore
                mood_template = MoodTemplate(lang=lang, mood_elem=mood_elem)
                self.mood_templates[mood_elem.tag.lower()] = mood_template
            self.modify_stem = ""
            modify_stem_attrib = template_elem.get("modify-stem", default=None)
            if modify_stem_attrib is not None:
                self.modify_stem = str(modify_stem_attrib)
                if self.modify_stem not in ("strip-accents"):
                    raise ConjugationTemplateError(
                        "Invalid 'modify-stem' attribute value '{self.modify_stem}'"
                    )
            else:
                self.modify_stem = ""

        except AttributeError as e:
            raise ConjugationTemplateError(
                "Error parsing {}: {}".format(etree.tostring(template_elem), str(e))
            )

    def __repr__(self) -> str:
        return "name={} mood_templates={}".format(self.name, self.mood_templates)



from enum import Enum


class AlternatesBehavior(Enum):
    FirstOnly = (
        1  # PersonConjugation is a scalar (str), using the first (default) conjugation
    )
    SecondOnly = 2  # PersonConjugation is a scalar (str), using the alternate conjugation (if available)
    All = 3  # PersonConjugation is a list (str) of all possible conjugations, default first



class ConjugatorError(Exception):
    pass


class ConjugationsParserError(Exception):
    pass


class ConjugationTemplateError(Exception):
    pass


class InvalidLangError(Exception):
    pass


class InvalidMoodError(Exception):
    pass


class InvalidTenseError(Exception):
    pass


class TemplateNotFoundError(Exception):
    pass


class VerbNotFoundError(Exception):
    pass


class VerbsParserError(Exception):
    pass



import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


# This refers to grammatical person ('usted' is 3s despite being semantically 2p)
class Person(StrEnum):
    FirstPersonSingular = "1s"
    SecondPersonSingular = "2s"
    ThirdPersonSingular = "3s"
    FirstPersonPlural = "1p"
    SecondPersonPlural = "2p"
    ThirdPersonPlural = "3p"


def is_plural(p: Person) -> bool:
    return p in (
        Person.FirstPersonPlural,
        Person.SecondPersonPlural,
        Person.ThirdPersonPlural,
    )


def is_singular(p: Person) -> bool:
    return not is_plural(p)



import logging

from verbecc.src.defs.constants.config import DEVEL_MODE
from verbecc.src.defs.types.mood import MoodEn as Mood
from verbecc.src.defs.types.tense import TenseEn as Tense
from verbecc.src.parsers.person_ending import PersonEnding

logging_level = logging.CRITICAL + 1  # effectively disables logging
if DEVEL_MODE:
    logging_level = logging.DEBUG

logging.basicConfig(
    level=logging_level,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    handlers=[logging.FileHandler("verbecc.log"), logging.StreamHandler()],
)

logger = logging.getLogger(__name__)

from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

from verbecc.src.defs.types.gender import Gender
from verbecc.src.defs.constants.grammar_defines import PARTICIPLE_INFLECTIONS
from verbecc.src.defs.types.exceptions import ConjugatorError
from verbecc.src.defs.types.lang_code import LangCodeISO639_1
from verbecc.src.defs.types.partiple_inflection import ParticipleInflection
from verbecc.src.defs.types.person import Person, is_singular
from verbecc.src.defs.types.lang_specific_options import (
    LangSpecificOptions,
)
from verbecc.src.parsers.conjugations_parser import ConjugationsParser
from verbecc.src.parsers.conjugation_template import ConjugationTemplate
from verbecc.src.parsers.tense_template import TenseTemplate
from verbecc.src.conjugator.conjugation_object import ConjugationObjects
from verbecc.src.parsers.verb import Verb
from verbecc.src.parsers.verbs_parser import VerbsParser


class Inflector(ABC):
    """
    Abstract base class, with an implementation for each specific language supported.
    Inflector is for all conjugation logic that is language-specific.
    """

    @property
    @abstractmethod
    def lang(self) -> LangCodeISO639_1:
        raise NotImplementedError

    def __init__(self) -> None:
        self._verb_parser = VerbsParser(self.lang)
        self._conj_parser = ConjugationsParser(self.lang)

    def get_verbs(self) -> List[Verb]:
        return self._verb_parser.verbs

    def get_infinitives(self) -> List[str]:
        return [v.infinitive for v in self._verb_parser.verbs]

    def get_templates(self) -> List[ConjugationTemplate]:
        return self._conj_parser.templates

    def get_template_names(self) -> List[str]:
        return [t.name for t in self._conj_parser.templates]

    def find_verb_by_infinitive(self, infinitive: str) -> Verb:
        return self._verb_parser.find_verb_by_infinitive(infinitive)

    def find_template(self, name: str) -> ConjugationTemplate:
        return self._conj_parser.find_template(name)

    def get_verbs_that_start_with(self, query: str, max_results: int) -> List[str]:
        query = query.lower()
        matches = self._verb_parser.get_verbs_that_start_with(query, max_results)
        return matches

    def _get_verb_stem_from_template_name(
        self, infinitive: str, template_name: str
    ) -> str:
        """Get the verb stem given an ininitive and a colon-delimited template name.
        E.g. infinitive='parler' template_name='aim:er' -> 'parl'
        Note: Catalan overrides this base class implementation to allow looser matching
        (only requires the last n-1 chars of template ending to match infinitive ending)
        """
        _, template_ending = template_name.split(":")
        if not infinitive.endswith(template_ending):
            raise ConjugatorError(
                "Template {} ending doesn't "
                "match infinitive {}".format(template_name, infinitive)
            )
        return infinitive[: len(infinitive) - len(template_ending)]

    def _is_impersonal_verb(self, infinitive: str) -> bool:
        return False

    def _verb_can_be_reflexive(self, infinitive: str) -> bool:
        return not self._is_impersonal_verb(infinitive)

    def _split_reflexive(self, infinitive: str) -> Tuple[bool, str]:
        """
        Tests whether an infinitive is reflexive
        Returns a 2-tuple of whether it is reflexive
        and the non-reflexive form of the infinitive.

        E.g. French:
        "se raser" => (True, "raser")
        "s'habiller" => (True, "habiller")
        "parler" => (False, "parler")
        E.g. Italian:
        "alzarsi" => (True, "alzare")
        "preoccuparsi" => (True, "preoccupare")
        """
        return (False, infinitive)

    def _add_reflexive_pronoun(self, s: str) -> str:
        return s

    def _add_subjunctive_relative_pronoun(self, s: str, tense: Tense) -> str:
        return s

    def _auxilary_verb_uses_alternate_conjugation(self, tense: Tense) -> bool:
        return False

    def _get_tenses_conjugated_without_pronouns(self) -> List[str]:
        return []

    def _get_auxilary_verb(
        self, co: ConjugationObjects, mood: Mood, tense: Tense
    ) -> str:
        return ""

    def _is_auxilary_verb_inflected(self, auxilary_verb: str) -> bool:
        return False

    def _get_infinitive_mood(self) -> Mood:
        return Mood.Infinitive

    def _get_indicative_mood(self) -> Mood:
        return Mood.Indicative

    def _get_subjunctive_mood(self) -> Mood:
        return Mood.Subjunctive

    def _get_conditional_mood(self) -> Mood:
        return Mood.Conditional

    def _get_participle_mood(self) -> Mood:
        return Mood.Participle

    def _get_participle_tense(self) -> Tense:
        return Tense.PastParticiple

    def _add_present_participle_if_applicable(
        self, s: str, is_reflexive: bool, tense: Tense
    ) -> str:
        return s

    def _get_alternate_hv_inflection(self, s: str) -> str:
        """Some language override this e.g. Spanish changes ending in 'hay' to 'ay'"""
        return s

    @abstractmethod
    def _get_compound_conjugations_aux_verb_map(
        self,
    ) -> Dict[str, Dict[str, Tuple[str, ...]]]:
        """Returns a map of the tense of the helping verb for each compound mood and tense"""
        raise NotImplementedError

    def _get_participle_index_for_participle_inflection(
        self, participle_inflection: ParticipleInflection
    ) -> int:
        """
        Default order is like French XML file, i.e. MS, MP, FS, FP
        But in some lang XML files, e.g. Italian, the order is MS, FS, MP, FP
        TODO: Standardize the XML files
        """
        return PARTICIPLE_INFLECTIONS.index(participle_inflection)

    def _get_default_participle_inflection_for_person(
        self, person: Person, gender: Gender = Gender.m
    ) -> ParticipleInflection:
        if is_singular(person):
            if gender == Gender.m:
                return ParticipleInflection.MasculineSingular
            else:
                return ParticipleInflection.FeminineSingular
        else:
            if gender == Gender.m:
                return ParticipleInflection.MasculinePlural
            else:
                return ParticipleInflection.FemininePlural

    def get_default_pronoun(
        self,
        person: Person,
        gender: Gender = Gender.m,
        is_reflexive: bool = False,
        lang_specific_options: LangSpecificOptions = None,
    ) -> str:
        return ""

    def _combine_pronoun_and_conj(self, pronoun: str, conj: str) -> str:
        return pronoun + " " + conj

    def _combine_verb_stem_and_ending(self, verb_stem: str, ending: str) -> str:
        """
        Originally this would simply combine the verb_stem and the ending.
        E.g. "parl" + "er" = "parler"

        Now for Catalan/enhanced templates we need to support stem-modifying verbs.

        E.g. for infintive "pertànyer" in the indicative mood simple-past tense,
        two new stem-modifying template feature will be used:

        First the template's modify-stem="strip-accents" attribute is used to
        remove the accent from the stem for all inflected forms.

        infinitive "pertànyer" matches template "pertàny:er" which has
        modify-stem="strip-accents"
        So the stem "pertàny" becomes "pertany"

        modify-stem="strip-accents" is applied in _conjugate_simple_mood_tense,
        so the verb_stem argument to this function is presumed to already have
        applicable modifications applied.

        Next, the "pertàny:er" template endings will use the new delete operator
        to delete one letter from the end of the stem.

        For "pertàny:er" in the i

        "pertany" + "-guí" = "pertanguí"

        Caution: A single "-" is also used as the placeholder for tenses that are
        not conjugated, in some verbs. "-" should only delete from the stem if it
        is followed by one or more characters.
        """
        while ending != "-" and ending.startswith("-"):
            ending = ending[1:]
            verb_stem = verb_stem[:-1]
        return verb_stem + ending

    def _get_pronoun_suffix(
        self, person: Person, gender: Gender = Gender.m, imperative: bool = True
    ) -> str:
        return " " + self.get_default_pronoun(person, gender)

    def _add_adverb_if_applicable(self, s: str, mood: Mood, tense: Tense) -> str:
        return s

    def _add_reflexive_pronoun_or_pronoun_suffix_if_applicable(
        self, s: str, is_reflexive: bool, mood: Mood, tense: Tense, person: Person
    ) -> str:
        if is_reflexive:
            s += self._get_pronoun_suffix(person)
        return s

    def _compound_conjugation_not_applicable(
        self, is_reflexive: bool, mood: Mood, aux_tense: Tense
    ) -> bool:
        return False

    def _compound_primary_verb_conjugation_uses_infinitive(
        self, mood: Mood, tense: Tense
    ) -> bool:
        return False

    def _modify_aux_verb_conj_if_applicable(
        self, aux_conj: List[str], mood: Mood, tense: Tense
    ) -> List[str]:
        """
        Hook for certain languages e.g. Romanian that use a different
        aux_conj for certain tenses
        Only used for Romanian conditional prezent and perfect, currently
        """
        return aux_conj

    def _add_compound_aux_verb_suffix_if_applicable(
        self, s: str, mood: Mood, tense: Mood
    ) -> str:
        """
        Hook for certain languages e.g. Romanian that add prefixes
        like ' fi' and 'să fi' for certain tenses
        Currently only used for certain Romanian verb tenses.
        """
        return s

    def _insert_compound_aux_verb_prefix_if_applicable(
        self, s: str, mood: Mood, tense: Tense
    ) -> str:
        """
        Used by Romanian viitor-1-popular
        "eu o să fac, tu o să faci, ..."
        """
        return s

    def _compound_has_no_primary_verb(self, mood: Mood, tense: Tense) -> bool:
        """Used for Romanian viitor-1-popular"""
        return False

    def _compound_has_no_aux_verb(self, mood: Mood, tense: Tense) -> bool:
        """Used for Romanian conjunctiv perfect"""
        return False

    def modify_person_ending_if_applicable(
        self,
        person_ending: PersonEnding,
        mood: Mood,
        tense: Tense,
        tense_template: TenseTemplate,
        lang_specific_options: LangSpecificOptions,
    ) -> PersonEnding:
        """
        Hook for certain languages e.g. Spanish that modify
        the standard person endings for certain persons depending
        on language specific options (e.g. Voseo)
        """
        return person_ending



from typing import Dict, List, Tuple

from verbecc.src.defs.types.gender import Gender
from verbecc.src.defs.types.lang_code import LangCodeISO639_1
from verbecc.src.defs.types.mood import MoodIt as Mood
from verbecc.src.defs.types.partiple_inflection import ParticipleInflection
from verbecc.src.defs.types.person import Person
from verbecc.src.defs.types.tense import TenseIt as Tense
from verbecc.src.defs.types.lang_specific_options import (
    LangSpecificOptions,
)
from verbecc.src.conjugator.conjugation_object import ConjugationObjects
from verbecc.src.inflectors.inflector import Inflector
from verbecc.src.utils import string_utils

VERBS_CONJUGATED_WITH_ESSERE = [
    "essere",
    "andare",
    "arrivare",
    "cadere",
    "entrare",
    "partire",
    "rimanere",
    "uscire",
    "venire",
    "stare",
    "passare",
    "diventare",
    "crescere",
    "morire",
    "nascere",
]


class InflectorIt(Inflector):
    @property
    def lang(self) -> LangCodeISO639_1:
        return LangCodeISO639_1.it

    def __init__(self) -> None:
        super(InflectorIt, self).__init__()

    def _is_auxilary_verb_inflected(self, auxilary_verb: str) -> bool:
        return auxilary_verb == "essere"

    def _split_reflexive(self, infinitive: str) -> Tuple[bool, str]:
        """
        E.g. Italian:
        "alzarsi" => (True, "alzare")
        "preoccuparsi" => (True, "preoccupare")

        TODO:
        Negative reflextive verbs:
        "non capire"
        """
        is_reflexive = False
        if infinitive.endswith("si"):
            is_reflexive = True
            infinitive = infinitive[:-2] + "e"  # "alzarsi" => "alzare"
        elif infinitive.startswith("si "):
            is_reflexive = True
            infinitive = infinitive[3:]
        elif infinitive.startswith("s'"):
            is_reflexive = True
            infinitive = infinitive[2:]
        return is_reflexive, infinitive

    def _add_reflexive_pronoun(self, s: str) -> str:
        if string_utils.starts_with_vowel(s, h_is_vowel=True):
            return "s'" + s
        else:
            return "si " + s

    def _add_subjunctive_relative_pronoun(self, s: str, tense: Tense) -> str:
        return "che " + s

    def get_default_pronoun(
        self,
        person: Person,
        gender: Gender = Gender.m,
        is_reflexive: bool = False,
        lang_specific_options: LangSpecificOptions = None,
    ) -> str:
        ret = ""
        if person == Person.FirstPersonSingular:
            ret = "io"
            if is_reflexive:
                ret += " mi"
        elif person == Person.SecondPersonSingular:
            ret = "tu"
            if is_reflexive:
                ret += " ti"
        elif person == Person.ThirdPersonSingular:
            ret = "lui"
            if gender == Gender.f:
                ret = "lei"
            if is_reflexive:
                ret += " si"
        elif person == Person.FirstPersonPlural:
            ret = "noi"
            if is_reflexive:
                ret += " ci"
        elif person == Person.SecondPersonPlural:
            ret = "voi"
            if is_reflexive:
                ret += " vi"
        elif person == Person.ThirdPersonPlural:
            ret = "loro"
            if is_reflexive:
                ret += " si"
        return ret

    def _get_tenses_conjugated_without_pronouns(self) -> List[Tense]:
        return [
            Tense.Affermativo,
            Tense.negativo,
            Tense.Negativo,
            Tense.Gerundio,
            Tense.ParticipioPresente,
            Tense.ParticipioPassato,
        ]

    def _get_auxilary_verb(
        self, co: ConjugationObjects, mood: Mood, tense: Tense
    ) -> str:
        ret = "avere"
        if co.verb.infinitive in VERBS_CONJUGATED_WITH_ESSERE or co.is_reflexive:
            ret = "essere"
        return ret

    def _get_infinitive_mood(self) -> Mood:
        return Mood.Infinito

    def _get_indicative_mood(self) -> Mood:
        return Mood.Indicativo

    def _get_subjunctive_mood(self) -> Mood:
        return Mood.Congiuntivo

    def _get_conditional_mood(self) -> Mood:
        return Mood.Condizionale

    def _get_participle_mood(self) -> Mood:
        return Mood.Participio

    def _get_participle_tense(self) -> Tense:
        return Tense.ParticipioPassato

    def _get_compound_conjugations_aux_verb_map(
        self,
    ) -> Dict[Mood, Dict[Tense, Tuple[Mood, Tense]]]:
        return {
            Mood.Indicativo: {
                Tense.PassatoProssimo: (Mood.Indicativo, Tense.Presente),
                Tense.TrapassatoProssimo: (Mood.Indicativo, Tense.Imperfetto),
                Tense.TrapassatoRemoto: (Mood.Indicativo, Tense.PassatoRemoto),
                Tense.FuturoAnteriore: (Mood.Indicativo, Tense.Futuro),
            },
            Mood.Congiuntivo: {
                Tense.Passato: (Mood.Congiuntivo, Tense.Presente),
                Tense.Trapassato: (Mood.Congiuntivo, Tense.Imperfetto),
            },
            Mood.Condizionale: {Tense.Passato: (Mood.Condizionale, Tense.Presente)},
        }

    PARTICIPLE_INFLECTIONS: Tuple[
        ParticipleInflection,
        ParticipleInflection,
        ParticipleInflection,
        ParticipleInflection,
    ] = (
        ParticipleInflection.MasculineSingular,
        ParticipleInflection.FeminineSingular,
        ParticipleInflection.MasculinePlural,
        ParticipleInflection.FemininePlural,
    )

    def _get_participle_index_for_participle_inflection(
        self, participle_inflection: ParticipleInflection
    ) -> int:
        """
        Default order is like French XML file, i.e. MS, MP, FS, FP
        But in some lang XML files, e.g. Italian, the order is MS, FS, MP, FP,
        so use the PARTICIPLE_INFLECTIONS above instead of the default one
        in grammar defines, for now.
        TODO: Standardize the XML files
        """
        return self.PARTICIPLE_INFLECTIONS.index(participle_inflection)



from unittest.mock import patch

import pytest

from verbecc.src.parsers.verbs_parser import VerbsParser
from verbecc.src.parsers.verb import Verb
from verbecc.src.defs.types.exceptions import VerbsParserError


def test_verbs_parser():
    vp = VerbsParser()
    assert len(vp.verbs) >= 7000


def test_verb():
    vp = VerbsParser()
    verb = vp.find_verb_by_infinitive("manger")
    assert verb.infinitive == "manger"
    assert verb.template == "man:ger"
    assert verb.translation_en == "eat"


def test_verb_two():
    vp = VerbsParser()
    verb = vp.find_verb_by_infinitive("abattre")
    assert verb.infinitive == "abattre"
    assert verb.template == "bat:tre"
    assert verb.translation_en == "tear down"


@patch("lxml.etree._Element")
def test_verb_invalid_xml(mock_v_elem):
    mock_v_elem.tag.return_value = "not-v"
    with pytest.raises(VerbsParserError):
        v = Verb(mock_v_elem)


@pytest.mark.parametrize(
    "query,expected_matches",
    [
        ("mang", ["mangeotter", "manger"]),
        ("Mang", ["mangeotter", "manger"]),
    ],
)
def test_get_verbs_that_start_with(query, expected_matches):
    vp = VerbsParser()
    matches = vp.get_verbs_that_start_with(query)
    assert matches == expected_matches



"""
Type annotations pytest test module.
Test to ensure all functions and methods have type annotations for parameters and return types.
This test scans the source code files in the specified directory, parses them using the `ast`
module, and checks each function and method definition for missing type annotations.
If any missing annotations are found, the test fails and reports the specific locations
and details of the missing annotations.
This test module utlizes the indirect parametrization feature of pytest to create a separate test case
for each missing type annotation found during the scan.
Copyright (c) 2026, Brett Tolbert <http://bretttolbert.com/>
"""

import pytest
import ast
import os
from typing import List

EXCLUDE_DIRS = {".git", "__pycache__", "venv", ".venv", "env", "build", "dist", "tests"}


class Finder(ast.NodeVisitor):
    def __init__(self, filename):
        self.filename = filename
        self.class_stack = []
        self.reported_errors: List[str] = []

    def visit_ClassDef(self, node):
        self.class_stack.append(node.name)
        self.generic_visit(node)
        self.class_stack.pop()

    def _report(self, node, func_name, missing_params, missing_return, lineno):
        parts = []
        if missing_params:
            parts.append("params: " + ", ".join(missing_params))
        if missing_return:
            parts.append("return")
        msg = f"{self.filename}:{lineno} - {'.'.join(self.class_stack + [func_name])} -> missing {', '.join(parts)}"
        print(msg)
        # assert False, msg
        self.reported_errors.append(msg)

    def _check_func(self, node):
        is_method = bool(self.class_stack)
        missing = []
        # posonlyargs (py3.8+), args, kwonlyargs
        all_args = []
        if hasattr(node.args, "posonlyargs"):
            all_args.extend(node.args.posonlyargs)
        all_args.extend(node.args.args)
        all_args.extend(node.args.kwonlyargs)

        for arg in all_args:
            # skip annotation check for typical self/cls on methods
            if is_method and arg.arg in ("self", "cls"):
                continue
            if arg.annotation is None:
                missing.append(arg.arg)

        # vararg / kwarg
        if node.args.vararg and node.args.vararg.annotation is None:
            missing.append("*" + node.args.vararg.arg)
        if node.args.kwarg and node.args.kwarg.annotation is None:
            missing.append("**" + node.args.kwarg.arg)

        missing_return = node.returns is None

        if missing or missing_return:
            self._report(node, node.name, missing, missing_return, node.lineno)

    def visit_FunctionDef(self, node):
        self._check_func(node)
        self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node):
        self._check_func(node)
        self.generic_visit(node)


def scan_file(path: str) -> List[str]:
    print("Scanning", path)
    try:
        with open(path, "r", encoding="utf-8") as f:
            src = f.read()
        tree = ast.parse(src, filename=path)
    except Exception as e:
        return [f"Path {path}: parse error: {e}"]
    finder = Finder(path)
    finder.visit(tree)
    return finder.reported_errors


def walk_root(root) -> List[str]:
    errors = []
    for dirpath, dirnames, filenames in os.walk(root):
        # skip excluded dirs
        dirnames[:] = [
            d
            for d in dirnames
            if d not in EXCLUDE_DIRS and not d.startswith(".egg-info")
        ]
        for fn in filenames:
            if fn.endswith(".py"):
                errors.extend(scan_file(os.path.join(dirpath, fn)))
    return errors


def pytest_generate_tests(metafunc):
    """Indirect parametrization for pytest to run test_per_error for each missing type annotation found."""
    if "errors" in metafunc.fixturenames:
        errors: List[str] = walk_root(metafunc.config.rootpath)
        metafunc.parametrize("errors", errors)


def test_per_error(errors: str):
    """Test that fails for each missing type annotation found.
    Works with pytest using indirect parametrization.
    (See pytest_generate_tests above.)"""
    error = errors
    assert False, error


if __name__ == "__main__":
    test_per_error()



import pytest

from verbecc.src.defs.types.gender import Gender
from verbecc.src.defs.types.person import Person
from verbecc.src.conjugator.conjugator import Conjugator
from verbecc.src.defs.types.alternates_behavior import AlternatesBehavior


@pytest.fixture(scope="module")
def cg():
    cg = Conjugator(lang="ro")
    yield cg


def test_all_verbs_have_templates(cg):
    verbs = cg.get_verbs()
    template_names = cg.get_template_names()
    missing_templates = set()
    for verb in verbs:
        if verb.template not in template_names:
            missing_templates.add(verb.template)
    assert len(missing_templates) == 0


@pytest.mark.parametrize(
    "infinitive,mood,tense,alternates_behavior,expected_result",
    [
        ("avea", "participiu", "participiu", AlternatesBehavior.FirstOnly, ["avut"]),
        ("face", "participiu", "participiu", AlternatesBehavior.FirstOnly, ["făcut"]),
        (
            "avea",
            "indicativ",
            "prezent",
            AlternatesBehavior.FirstOnly,
            ["eu am", "tu ai", "el a", "noi am", "voi aţi", "ei au"],
        ),
        (
            "avea",
            "indicativ",
            "prezent",
            AlternatesBehavior.SecondOnly,
            ["eu am", "tu ai", "el are", "noi avem", "voi aveţi", "ei au"],
        ),
        (
            "avea",
            "indicativ",
            "imperfect",
            AlternatesBehavior.FirstOnly,
            ["eu aveam", "tu aveai", "el avea", "noi aveam", "voi aveaţi", "ei aveau"],
        ),
        (
            "avea",
            "indicativ",
            "perfect-simplu",
            AlternatesBehavior.FirstOnly,
            ["eu avui", "tu avuși", "el avu", "noi avurăm", "voi avurăţi", "ei avură"],
        ),
        (
            "avea",
            "indicativ",
            "perfect-compus",
            AlternatesBehavior.FirstOnly,
            [
                "eu am avut",
                "tu ai avut",
                "el a avut",
                "noi am avut",
                "voi aţi avut",
                "ei au avut",
            ],
        ),
        (
            "avea",
            "indicativ",
            "mai-mult-ca-perfect",
            AlternatesBehavior.FirstOnly,
            [
                "eu avusem",
                "tu avuseși",
                "el avuse",
                "noi avuserăm",
                "voi avuserăţi",
                "ei avuseră",
            ],
        ),
        (
            "face",
            "indicativ",
            "prezent",
            AlternatesBehavior.FirstOnly,
            ["eu fac", "tu faci", "el face", "noi facem", "voi faceţi", "ei fac"],
        ),
        (
            "face",
            "indicativ",
            "imperfect",
            AlternatesBehavior.FirstOnly,
            [
                "eu făceam",
                "tu făceai",
                "el făcea",
                "noi făceam",
                "voi făceaţi",
                "ei făceau",
            ],
        ),
        (
            "face",
            "indicativ",
            "perfect-simplu",
            AlternatesBehavior.FirstOnly,
            [
                "eu făcui",
                "tu făcuși",
                "el făcu",
                "noi făcurăm",
                "voi făcurăţi",
                "ei făcură",
            ],
        ),
        (
            "face",
            "indicativ",
            "perfect-compus",
            AlternatesBehavior.FirstOnly,
            [
                "eu am făcut",
                "tu ai făcut",
                "el a făcut",
                "noi am făcut",
                "voi aţi făcut",
                "ei au făcut",
            ],
        ),
        (
            "face",
            "indicativ",
            "mai-mult-ca-perfect",
            AlternatesBehavior.FirstOnly,
            [
                "eu făcusem",
                "tu făcuseși",
                "el făcuse",
                "noi făcuserăm",
                "voi făcuserăţi",
                "ei făcuseră",
            ],
        ),
        (
            "voi",
            "indicativ",
            "prezent",
            AlternatesBehavior.FirstOnly,
            [
                "eu voiesc",
                "tu voiești",
                "el voiește",
                "noi voim",
                "voi voiţi",
                "ei voiesc",
            ],
        ),
        (
            "voi",
            "indicativ",
            "prezent",
            AlternatesBehavior.SecondOnly,
            ["eu voi", "tu vei", "el va", "noi vom", "voi veţi", "ei vor"],
        ),
        (
            "face",
            "indicativ",
            "viitor-1",
            AlternatesBehavior.FirstOnly,
            [
                "eu voi face",
                "tu vei face",
                "el va face",
                "noi vom face",
                "voi veţi face",
                "ei vor face",
            ],
        ),
        (
            "face",
            "indicativ",
            "viitor-2",
            AlternatesBehavior.FirstOnly,
            [
                "eu voi fi făcut",
                "tu vei fi făcut",
                "el va fi făcut",
                "noi vom fi făcut",
                "voi veţi fi făcut",
                "ei vor fi făcut",
            ],
        ),
        (
            "face",
            "conjunctiv",
            "prezent",
            AlternatesBehavior.FirstOnly,
            [
                "eu să fac",
                "tu să faci",
                "el să facă",
                "noi să facem",
                "voi să faceţi",
                "ei să facă",
            ],
        ),
        (
            "face",
            "conjunctiv",
            "perfect",
            AlternatesBehavior.FirstOnly,
            [
                "eu să fi făcut",
                "tu să fi făcut",
                "el să fi făcut",
                "noi să fi făcut",
                "voi să fi făcut",
                "ei să fi făcut",
            ],
        ),
        (
            "face",
            "indicativ",
            "viitor-1-popular",
            AlternatesBehavior.FirstOnly,
            [
                "eu o să fac",
                "tu o să faci",
                "el o să facă",
                "noi o să facem",
                "voi o să faceţi",
                "ei o să facă",
            ],
        ),
        (
            "face",
            "indicativ",
            "viitor-2-popular",
            AlternatesBehavior.FirstOnly,
            [
                "eu am să fi făcut",
                "tu ai să fi făcut",
                "el are să fi făcut",
                "noi avem să fi făcut",
                "voi aveţi să fi făcut",
                "ei au să fi făcut",
            ],
        ),
        (
            "avea",
            "conjunctiv",
            "prezent",
            AlternatesBehavior.FirstOnly,
            [
                "eu să am",
                "tu să ai",
                "el să aibă",
                "noi să avem",
                "voi să aveţi",
                "ei să aibă",
            ],
        ),
        (
            "avea",
            "condițional",
            "prezent",
            AlternatesBehavior.FirstOnly,
            [
                "eu aş avea",
                "tu ai avea",
                "el ar avea",
                "noi am avea",
                "voi aţi avea",
                "ei ar avea",
            ],
        ),
        (
            "avea",
            "condițional",
            "perfect",
            AlternatesBehavior.FirstOnly,
            [
                "eu aş fi avut",
                "tu ai fi avut",
                "el ar fi avut",
                "noi am fi avut",
                "voi aţi fi avut",
                "ei ar fi avut",
            ],
        ),
    ],
)
def test_inflector_ro_conjugate_mood_tense(
    cg, infinitive, mood, tense, alternates_behavior, expected_result
):
    assert (
        cg.conjugate_mood_tense(infinitive, mood, tense, alternates_behavior)
        == expected_result
    )


def test_inflector_ro_conjugate_mood_tense_viitor_1(cg):
    assert cg.conjugate_mood_tense(
        "face", "indicativ", "viitor-1", AlternatesBehavior.FirstOnly
    ) == [
        "eu voi face",
        "tu vei face",
        "el va face",
        "noi vom face",
        "voi veţi face",
        "ei vor face",
    ]


def test_inflector_ro_conjugate_mood_tense_viitor_1_popular(cg):
    assert cg.conjugate_mood_tense(
        "face",
        "indicativ",
        "viitor-1-popular",
        AlternatesBehavior.FirstOnly,
        [
            "eu o să fac",
            "tu o să faci",
            "el o să facă",
            "noi o să facem",
            "voi o să faceţi",
            "ei o să facă",
        ],
    )


def test_inflector_ro_conjugate_mood_tense_condițional_perfect(cg):
    assert cg.conjugate_mood_tense(
        "avea", "condițional", "perfect", AlternatesBehavior.FirstOnly
    ) == [
        "eu aş fi avut",
        "tu ai fi avut",
        "el ar fi avut",
        "noi am fi avut",
        "voi aţi fi avut",
        "ei ar fi avut",
    ]


def test_inflector_ro_conjugate_mood_tense_conjunctiv_perfect(cg):
    assert cg.conjugate_mood_tense(
        "face", "conjunctiv", "perfect", AlternatesBehavior.FirstOnly
    ) == [
        "eu să fi făcut",
        "tu să fi făcut",
        "el să fi făcut",
        "noi să fi făcut",
        "voi să fi făcut",
        "ei să fi făcut",
    ]


@pytest.mark.parametrize(
    "person,gender,is_reflexive,expected_result",
    [
        (Person.FirstPersonSingular, Gender.m, False, "eu"),
        (Person.FirstPersonSingular, Gender.m, True, "eu mă"),
        (Person.SecondPersonSingular, Gender.m, False, "tu"),
        (Person.SecondPersonSingular, Gender.m, True, "tu te"),
        (Person.ThirdPersonSingular, Gender.m, False, "el"),
        (Person.ThirdPersonSingular, Gender.m, True, "el se"),
        (Person.ThirdPersonSingular, Gender.f, False, "ea"),
        (Person.ThirdPersonSingular, Gender.f, True, "ea se"),
        (Person.FirstPersonPlural, Gender.m, False, "noi"),
        (Person.FirstPersonPlural, Gender.m, True, "noi ne"),
        (Person.SecondPersonPlural, Gender.m, False, "voi"),
        (Person.SecondPersonPlural, Gender.m, True, "voi vă"),
        (Person.ThirdPersonPlural, Gender.m, False, "ei"),
        (Person.ThirdPersonPlural, Gender.m, True, "ei se"),
        (Person.ThirdPersonPlural, Gender.f, False, "ele"),
        (Person.ThirdPersonPlural, Gender.f, True, "ele se"),
    ],
)
def test_inflector_ro_get_default_pronoun(
    cg, person: Person, gender: Gender, is_reflexive: bool, expected_result: str
):
    assert (
        cg._inflector.get_default_pronoun(person, gender, is_reflexive=is_reflexive)
        == expected_result
    )
